"""
$description Live TV channels from LTV, a Latvian public, state-owned broadcaster.
$url ltv.lsm.lv
$url replay.lsm.lv
$type live, vod
$region Latvia
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https://(?:ltv|replay)\.lsm\.lv/(?:lv/tiesraide|ru/efir)/"),
)
class LtvLsmLv(Plugin):
    URL_API = "https://player.cloudycdn.services/player/ltvlive/channel/{channel_id}/"

    def _get_streams(self):
        self.session.http.headers.update({"Referer": self.url})

        iframe_url = self.session.http.get(
            self.url,
            schema=validate.Schema(
                re.compile(r"""(?P<q>")https:\\u002F\\u002Fltv\.lsm\.lv\\u002Fembed\\u002Flive\?\S+?(?P=q)"""),
                validate.none_or_all(
                    validate.get(0),
                    validate.parse_json(),
                ),
            ),
        )
        if not iframe_url:
            log.error("Could not find video player iframe")
            return

        starts_at, channel_id = self.session.http.get(
            iframe_url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//live[1]/@*[name()=':embed-data']"),
                str,
                validate.parse_json(),
                {
                    "parentInfo": {"starts_at": validate.any(None, str)},
                    "source": {"item_id": str},
                },
                validate.union_get(
                    ("parentInfo", "starts_at"),
                    ("source", "item_id"),
                ),
            ),
        )
        if channel_id is None:
            return
        log.debug(f"Found channel ID: {channel_id}")

        if starts_at is not None:
            log.error(f"Stream starts at {starts_at}")
            return

        stream_sources = self.session.http.post(
            self.URL_API.format(channel_id=channel_id),
            data={
                "refer": "ltv.lsm.lv",
                "playertype": "regular",
                "protocol": "hls",
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "source": {
                        "sources": validate.all(
                            [{"type": str, "src": validate.url()}],
                            validate.filter(lambda src: src["type"] == "application/x-mpegURL"),
                            validate.map(lambda src: src.get("src")),
                        ),
                    },
                },
                validate.get(("source", "sources")),
            ),
        )
        for surl in stream_sources:
            yield from HLSStream.parse_variant_playlist(self.session, surl).items()


__plugin__ = LtvLsmLv



from __future__ import annotations

import tempfile
from contextlib import asynccontextmanager, contextmanager
from functools import partial
from subprocess import DEVNULL
from typing import TYPE_CHECKING

import trio

from streamlink.compat import BaseExceptionGroup
from streamlink.logger import getLogger
from streamlink.utils.path import resolve_executable
from streamlink.webbrowser.exceptions import WebbrowserError


if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Generator
    from contextlib import AbstractAsyncContextManager
    from pathlib import Path


log = getLogger(__name__)


class Webbrowser:
    ERROR_RESOLVE = "Could not find web browser executable"

    TIMEOUT = 10

    @classmethod
    def names(cls) -> list[str]:
        return []

    @classmethod
    def fallback_paths(cls) -> list[str | Path]:
        return []

    @classmethod
    def launch_args(cls) -> list[str]:
        return []

    def __init__(self, executable: str | None = None):
        resolved = resolve_executable(executable, self.names(), self.fallback_paths())
        if not resolved:
            raise WebbrowserError(
                f"Invalid web browser executable: {executable}"
                if executable
                else f"{self.ERROR_RESOLVE}: Please set the path to a supported web browser using --webbrowser-executable",
            )

        self.executable: str | Path = resolved
        self.arguments: list[str] = self.launch_args().copy()

    def launch(self, headless: bool = False, timeout: float | None = None) -> AbstractAsyncContextManager[trio.Nursery]:
        return self._launch(self.executable, self.arguments, headless=headless, timeout=timeout)

    def _launch(
        self,
        executable: str | Path,
        arguments: list[str],
        headless: bool = False,
        timeout: float | None = None,
    ) -> AbstractAsyncContextManager[trio.Nursery]:
        if timeout is None:
            timeout = self.TIMEOUT

        launcher = _WebbrowserLauncher(executable, arguments, headless, timeout)

        # noinspection PyArgumentList
        return launcher.launch()

    @staticmethod
    @contextmanager
    def _create_temp_dir() -> Generator[str, None, None]:
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as temp_file:
            yield temp_file


class _WebbrowserLauncher:
    def __init__(self, executable: str | Path, arguments: list[str], headless: bool, timeout: float):
        self.executable = executable
        self.arguments = arguments
        self.headless = headless
        self.timeout = timeout
        self._process_ended_early = False

    @asynccontextmanager
    async def launch(self) -> AsyncGenerator[trio.Nursery, None]:
        try:
            headless = self.headless
            async with trio.open_nursery() as nursery:
                log.info(f"Launching web browser: {self.executable} ({headless=})")
                # the process is run in a separate task
                run_process = partial(
                    trio.run_process,
                    [self.executable, *self.arguments],
                    check=False,
                    stdout=DEVNULL,
                    stderr=DEVNULL,
                )
                # trio ensures that the process gets terminated when the task group gets cancelled
                process: trio.Process = await nursery.start(run_process)
                # the process watcher task cancels the entire task group when the user terminates/kills the process
                nursery.start_soon(self._task_process_watcher, process, nursery)
                try:
                    # the application logic is run here
                    with trio.move_on_after(self.timeout) as cancel_scope:
                        yield nursery
                    # check if the application logic has timed out
                    if cancel_scope.cancelled_caught:
                        log.warning("Web browser task group has timed out")
                finally:
                    # check if the task group hasn't been cancelled yet in the process watcher task
                    if not self._process_ended_early:
                        log.debug("Waiting for web browser process to terminate")
                    # once the application logic is done, cancel the entire task group and terminate/kill the process
                    nursery.cancel_scope.cancel()
        except BaseExceptionGroup as exc_grp:  # TODO: py310 support end: use except*
            exc: BaseException | BaseExceptionGroup | None = exc_grp.subgroup((KeyboardInterrupt, SystemExit))
            if not exc:  # not a KeyboardInterrupt or SystemExit
                raise
            while isinstance(exc, BaseExceptionGroup):  # get the first actual exception in the potentially nested groups
                exc = exc.exceptions[0]
            raise exc from exc.__context__

    async def _task_process_watcher(self, process: trio.Process, nursery: trio.Nursery) -> None:
        """Task for cancelling the launch task group if the user closes the browser or if it exits early on its own"""
        await process.wait()
        # if the task group hasn't been cancelled yet, then the application logic was still running
        if not nursery.cancel_scope.cancel_called:  # pragma: no branch
            self._process_ended_early = True
            log.warning("Web browser process ended early")
            nursery.cancel_scope.cancel()



from __future__ import annotations

import os
from collections import deque
from math import floor
from string import Formatter as StringFormatter
from threading import Event, RLock, Thread
from time import time
from typing import TYPE_CHECKING, TypeAlias

from streamlink_cli.console.terminal import cut_text, term_width, text_width


if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Mapping
    from pathlib import PurePath

    from streamlink_cli.console.console import ConsoleOutput

    _TFormat: TypeAlias = Iterable[Iterable[tuple[str, str | None, str | None, str | None]]]


_stringformatter = StringFormatter()


class ProgressFormatter:
    # Store formats as a tuple of lists of parsed format strings,
    # so when iterating, we don't have to parse over and over again.
    # Reserve at least 15 characters for the path, so it can be truncated with enough useful information.
    FORMATS: _TFormat = tuple(
        list(_stringformatter.parse(fmt))
        for fmt in (
            "[download] Written {written} to {path:15} ({elapsed} @ {speed})",
            "[download] Written {written} ({elapsed} @ {speed})",
            "[download] {written} ({elapsed} @ {speed})",
            "[download] {written} ({elapsed})",
            "[download] {written}",
        )
    )
    FORMATS_NOSPEED: _TFormat = tuple(
        list(_stringformatter.parse(fmt))
        for fmt in (
            "[download] Written {written} to {path:15} ({elapsed})",
            "[download] Written {written} ({elapsed})",
            "[download] {written} ({elapsed})",
            "[download] {written}",
        )
    )

    # Use U+2026 (HORIZONTAL ELLIPSIS) to be able to distinguish between "." and ".." when truncating relative paths
    ELLIPSIS: str = "…"

    @classmethod
    def format(cls, formats: _TFormat, params: Mapping[str, str | Callable[[int], str]]) -> str:
        width = term_width()
        static: list[str] = []
        variable: list[tuple[int, Callable[[int], str], int]] = []

        for fmt in formats:
            static.clear()
            variable.clear()
            length = 0
            # Get literal texts, static segments and variable segments from the parsed format
            # and calculate the overall length of the literal texts and static segments after substituting them.
            for literal_text, field_name, format_spec, _conversion in fmt:
                static.append(literal_text)
                length += len(literal_text)
                if field_name is None:
                    continue
                if field_name not in params:
                    break
                value_or_callable = params[field_name]
                if isinstance(value_or_callable, str):
                    static.append(value_or_callable)
                    length += len(value_or_callable)
                else:
                    variable.append((len(static), value_or_callable, int(format_spec or 0)))
                    static.append("")
            else:
                # No variable segments? Just check if the resulting string fits into the size constraints.
                if not variable:
                    if length > width:
                        continue
                    else:
                        break

                # Get the available space for each variable segment (share space equally and round down).
                max_width = int((width - length) / len(variable))
                # If at least one variable segment doesn't fit, continue with the next format.
                if max_width < 1 or any(max_width < min_width for _, __, min_width in variable):
                    continue
                # All variable segments fit, so finally format them, but continue with the next format if there's an error.
                # noinspection PyBroadException
                try:
                    for idx, fn, _ in variable:
                        static[idx] = fn(max_width)
                except Exception:
                    continue
                break

        return "".join(static)

    @staticmethod
    def _round(num: float, n: int = 2) -> float:
        return floor(num * 10**n) / 10**n

    @classmethod
    def format_filesize(cls, size: float, suffix: str = "") -> str:
        if size < 1024:
            return f"{size:.0f} bytes{suffix}"
        if size < 2**20:
            return f"{cls._round(size / 2**10, 2):.2f} KiB{suffix}"
        if size < 2**30:
            return f"{cls._round(size / 2**20, 2):.2f} MiB{suffix}"
        if size < 2**40:
            return f"{cls._round(size / 2**30, 2):.2f} GiB{suffix}"

        return f"{cls._round(size / 2**40, 2):.2f} TiB{suffix}"

    @classmethod
    def format_time(cls, elapsed: float) -> str:
        elapsed = max(elapsed, 0)

        if elapsed < 60:
            return f"{int(elapsed % 60):1d}s"
        if elapsed < 3600:
            return f"{int(elapsed % 3600 / 60):1d}m{int(elapsed % 60):02d}s"

        return f"{int(elapsed / 3600)}h{int(elapsed % 3600 / 60):02d}m{int(elapsed % 60):02d}s"

    @classmethod
    def format_path(cls, path: PurePath, max_width: int) -> str:
        # Quick check if the path fits
        string = str(path)
        width = text_width(string)
        if width <= max_width:
            return string

        # Since the path doesn't fit, we always need to add an ellipsis.
        # On Windows, we also need to add the "drive" part (which is an empty string on PurePosixPath)
        max_width -= text_width(path.drive) + text_width(cls.ELLIPSIS)

        # Ignore the path's first part, aka the "anchor" (drive + root)
        parts = os.path.sep.join(path.parts[1:] if path.drive else path.parts)
        truncated = cut_text(parts, max_width)

        return f"{path.drive}{cls.ELLIPSIS}{truncated}"


class Progress(Thread):
    def __init__(
        self,
        console: ConsoleOutput,
        path: PurePath,
        interval: float = 0.25,
        history: int = 20,
        threshold: int = 2,
        status: bool = True,
    ):
        """
        :param console: The console output
        :param interval: Time in seconds between updates
        :param history: Number of seconds of how long download speed history is kept
        :param threshold: Number of seconds until download speed is shown
        """

        super().__init__(daemon=True)
        self._wait = Event()
        self._lock = RLock()

        self.formatter = ProgressFormatter()

        self.console: ConsoleOutput = console
        self.path: PurePath = path
        self.interval: float = interval
        self.history: deque[tuple[float, int]] = deque(maxlen=int(history / interval))
        self.threshold: int = int(threshold / interval)

        self.started: float = 0.0
        self.overall: int = 0
        self.written: int = 0
        self.status: bool = status

    def close(self):
        self._wait.set()

    def write(self, chunk: bytes):
        size = len(chunk)
        with self._lock:
            self.overall += size
            self.written += size

    def run(self):
        self.started = time()
        try:
            while not self._wait.wait(self.interval):
                self.update()
        finally:
            self.update()

    def update(self):
        with self._lock:
            now = time()
            formatter = self.formatter
            history = self.history

            history.append((now, self.written))
            self.written = 0

            has_history = len(history) >= self.threshold
            if not has_history or now == history[0][0]:
                formats = formatter.FORMATS_NOSPEED
                speed = ""
            else:
                formats = formatter.FORMATS
                speed = formatter.format_filesize(sum(size for _, size in history) / (now - history[0][0]), "/s")

            params = dict(
                written=formatter.format_filesize(self.overall),
                elapsed=formatter.format_time(now - self.started),
                speed=speed,
                path=lambda max_width: formatter.format_path(self.path, max_width),
            )

            status = formatter.format(formats, params)
            if self.status:
                self.console.msg_status(status)
            else:
                self.console.msg(status)



"""
$description Chinese live-streaming platform owned by ByteDance.
$url live.douyin.com
$type live
$metadata id
$metadata author
$metadata title
"""

from __future__ import annotations

import re
import uuid

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.http import HTTPStream
from streamlink.utils.url import update_scheme


log = getLogger(__name__)


@pluginmatcher(
    re.compile(
        r"https?://(?:live\.)?douyin\.com/(?P<room_id>[^/?]+)",
    ),
)
class Douyin(Plugin):
    _STATUS_LIVE = 2

    QUALITY_WEIGHTS: dict[str, int] = {}

    @classmethod
    def stream_weight(cls, stream: str) -> tuple[float, str]:
        weight = cls.QUALITY_WEIGHTS.get(stream)
        if weight:
            return weight, stream

        return super().stream_weight(stream)

    SCHEMA_ROOM_STORE = validate.all(
        {
            "roomInfo": {
                # "room" and "anchor" keys are missing on invalid channels
                validate.optional("room"): validate.all(
                    {
                        "id_str": str,
                        "status": int,
                        "title": str,
                    },
                    validate.union_get(
                        "status",
                        "id_str",
                        "title",
                    ),
                ),
                validate.optional("anchor"): validate.all(
                    {"nickname": str},
                    validate.get("nickname"),
                ),
            },
        },
        validate.union_get(
            ("roomInfo", "room"),
            ("roomInfo", "anchor"),
        ),
    )

    SCHEMA_STREAM_STORE = validate.all(
        {
            "streamData": {
                "H264_streamData": {
                    # "stream" value is `none` on offline/invalid channels
                    "stream": validate.none_or_all(
                        {
                            str: validate.all(
                                {
                                    "main": {
                                        # HLS stream URLs are multivariant streams but only with a single media playlist,
                                        # so avoid using HLS in favor of having reduced stream lookup/start times
                                        "flv": validate.any("", validate.url()),
                                        "sdk_params": validate.all(
                                            validate.parse_json(),
                                            {"vbitrate": int},
                                            validate.get("vbitrate"),
                                        ),
                                    },
                                },
                                validate.union_get(
                                    ("main", "sdk_params"),
                                    ("main", "flv"),
                                ),
                            ),
                        },
                    ),
                },
            },
        },
        validate.get(("streamData", "H264_streamData", "stream")),
    )

    def _get_streams(self):
        data = self.session.http.get(
            self.url,
            cookies={
                "__ac_nonce": uuid.uuid4().hex[:21],
            },
            schema=validate.Schema(
                validate.regex(
                    pattern=re.compile(r"self\.__pace_f\.push\(\[\d+,(\"\w+:.+?\")]\)</script>"),
                    method="findall",
                ),
                validate.filter(lambda item: "state" in item and "streamStore" in item),
                validate.get(-1),
                validate.none_or_all(
                    validate.parse_json(),
                    validate.transform(lambda s: re.sub(r"^\w+:", "", s)),
                    validate.parse_json(),
                    list,
                    validate.filter(lambda item: isinstance(item, dict) and "state" in item),
                    validate.length(1),
                    validate.get((0, "state")),
                    {
                        "roomStore": self.SCHEMA_ROOM_STORE,
                        "streamStore": self.SCHEMA_STREAM_STORE,
                    },
                    validate.union_get(
                        "roomStore",
                        "streamStore",
                    ),
                ),
            ),
        )
        if not data:
            return

        (room_info, self.author), stream_data = data
        if not room_info:
            return

        status, self.id, self.title = room_info
        if status != self._STATUS_LIVE:
            log.info("The channel is currently offline")
            return

        for name, (vbitrate, url) in stream_data.items():
            if not url:
                continue
            self.QUALITY_WEIGHTS[name] = vbitrate
            url = update_scheme("https://", url, force=True)
            yield name, HTTPStream(self.session, url)

        log.debug("self.QUALITY_WEIGHTS=%r", self.QUALITY_WEIGHTS)


__plugin__ = Douyin



"""
Streamlink extracts streams from various services.

The main compontent of Streamlink is a command-line utility that
launches the streams in a video player.

An API is also provided that allows direct access to stream data.

Full documentation is available at https://streamlink.github.io/
"""

from streamlink._version import __version__
from streamlink.api import streams
from streamlink.exceptions import StreamlinkError, PluginError, NoStreamsError, NoPluginError, StreamError
from streamlink.session import Streamlink


__title__ = "streamlink"
__license__ = "Simplified BSD"
__author__ = "Streamlink"
__copyright__ = "Copyright 2026 Streamlink"
__credits__ = ["https://github.com/streamlink/streamlink/blob/master/AUTHORS"]



"""
$description Turkish live TV channel owned by Disney.
$url nowtv.com.tr
$type live, vod
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?nowtv\.com\.tr/"),
)
class NowTVTR(Plugin):
    def _get_streams(self):
        stream_url = self.session.http.get(
            self.url,
            schema=validate.Schema(
                re.compile(r"""(?P<q>['"])(?P<url>https://nowtv[^/]+/\S+/playlist\.m3u8\?\S+)(?P=q)"""),
                validate.none_or_all(validate.get("url")),
            ),
        )
        if stream_url:
            return HLSStream.parse_variant_playlist(self.session, stream_url)


__plugin__ = NowTVTR



# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all modules.
#
# CDP version: v0.0.1510116
# CDP domain: IO

from __future__ import annotations

import enum
from collections.abc import Generator
from dataclasses import dataclass
from typing import Any

import streamlink.webbrowser.cdp.devtools.runtime as runtime
from streamlink.webbrowser.cdp.devtools.util import T_JSON_DICT, CDPEvent


class StreamHandle(str):
    """
    This is either obtained from another method or specified as ``blob:<uuid>`` where
    ``<uuid>`` is an UUID of a Blob.
    """
    def to_json(self) -> str:
        return self

    @classmethod
    def from_json(cls, json: str) -> StreamHandle:
        return cls(json)

    def __repr__(self):
        return f"StreamHandle({super().__repr__()})"


def close(
    handle: StreamHandle,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Close the stream, discard any temporary backing storage.

    :param handle: Handle of the stream to close.
    """
    params: T_JSON_DICT = {}
    params["handle"] = handle.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "IO.close",
        "params": params,
    }
    yield cmd_dict


def read(
    handle: StreamHandle,
    offset: int | None = None,
    size: int | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, tuple[bool | None, str, bool]]:
    """
    Read a chunk of the stream

    :param handle: Handle of the stream to read.
    :param offset: *(Optional)* Seek to the specified offset before reading (if not specified, proceed with offset following the last read). Some types of streams may only support sequential reads.
    :param size: *(Optional)* Maximum number of bytes to read (left upon the agent discretion if not specified).
    :returns: A tuple with the following items:

        0. **base64Encoded** - *(Optional)* Set if the data is base64-encoded
        1. **data** - Data that were read.
        2. **eof** - Set if the end-of-file condition occurred while reading.
    """
    params: T_JSON_DICT = {}
    params["handle"] = handle.to_json()
    if offset is not None:
        params["offset"] = offset
    if size is not None:
        params["size"] = size
    cmd_dict: T_JSON_DICT = {
        "method": "IO.read",
        "params": params,
    }
    json = yield cmd_dict
    return (
        bool(json["base64Encoded"]) if "base64Encoded" in json else None,
        str(json["data"]),
        bool(json["eof"]),
    )


def resolve_blob(
    object_id: runtime.RemoteObjectId,
) -> Generator[T_JSON_DICT, T_JSON_DICT, str]:
    """
    Return UUID of Blob object specified by a remote object id.

    :param object_id: Object id of a Blob object wrapper.
    :returns: UUID of the specified Blob.
    """
    params: T_JSON_DICT = {}
    params["objectId"] = object_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "IO.resolveBlob",
        "params": params,
    }
    json = yield cmd_dict
    return str(json["uuid"])



from __future__ import annotations

from typing import Any


def search_dict(data: dict | list, key: Any):
    """
    Search for a key in a nested dict, or list of nested dicts, and return the values.

    :param data: dict/list to search
    :param key: key to find
    :return: matches for key
    """
    if isinstance(data, dict):
        for dkey, value in data.items():
            if dkey == key:
                yield value
            yield from search_dict(value, key)
    elif isinstance(data, list):
        for value in data:
            yield from search_dict(value, key)



from __future__ import annotations

import importlib
import inspect
import os
import sys
import warnings
from typing import TYPE_CHECKING, Any


try:
    from builtins import BaseExceptionGroup, ExceptionGroup  # type: ignore[attr-defined, ty:unresolved-import]
except ImportError:  # pragma: no cover
    from exceptiongroup import BaseExceptionGroup, ExceptionGroup  # type: ignore[import, ty:unresolved-import]

from requests.compat import chardet as charset_normalizer  # type: ignore[attr-defined, ty:unresolved-import]

from streamlink.exceptions import StreamlinkDeprecationWarning


if TYPE_CHECKING:
    from collections.abc import Callable, Mapping
    from typing import Protocol, TypedDict

    # The return value of requests' compatibility import function for charset_normalizer/chardet can be None,
    # even though it's not possible due to charset_normalizer being a dependency since requests 2.26.0,
    # with chardet being an optional dependency (that is still prioritized over charset_normalizer though).
    # Make type checkers happy with this assertion once requests switches to inline typing annotations.
    # With the requests type stubs provided by typeshed, we suppress the import type error.
    assert charset_normalizer is not None

    class _DetectEncodingResult(TypedDict):
        encoding: str | None
        language: str
        confidence: float | None

    class _DetectEncoding(Protocol):
        def __call__(self, byte_str: bytes, should_rename_legacy: bool = False, **kwargs: Any) -> _DetectEncodingResult: ...


is_linux = sys.platform == "linux"
is_android = sys.platform == "android"
is_darwin = sys.platform == "darwin"
is_freebsd = sys.platform.startswith("freebsd")
is_posix = os.name == "posix"
is_win32 = os.name == "nt"


detect_encoding: _DetectEncoding = charset_normalizer.detect


def deprecated(items: Mapping[str, tuple[str | None, Any, Any]]) -> None:
    """
    Deprecate specific module attributes.

    This removes the deprecated attributes from the module's global context,
    adds/overrides the module's :func:`__getattr__` function, and emits a :class:`StreamlinkDeprecationWarning`
    if one of the deprecated attributes is accessed.

    :param items: A mapping of module attribute names to tuples which contain the following optional items:
                  1. an import path string (for looking up an external object while accessing the attribute)
                  2. a direct return object (if no import path was set)
                  3. a custom warning message
    """

    mod_globals = inspect.stack()[1].frame.f_globals
    orig_getattr: Callable[[str], Any] | None = mod_globals.get("__getattr__", None)

    def __getattr__(name: str) -> Any:
        if name in items:
            origin = f"{mod_globals['__spec__'].name}.{name}"
            path, obj, msg = items[name]
            warnings.warn(
                msg or f"'{origin}' has been deprecated",
                StreamlinkDeprecationWarning,
                stacklevel=2,
            )
            if path:
                *parts, name = path.split(".")
                imported = importlib.import_module(".".join(parts))
                obj = getattr(imported, name, None)

            return obj

        if orig_getattr is not None:
            return orig_getattr(name)

        raise AttributeError

    mod_globals["__getattr__"] = __getattr__

    # delete the deprecated module attributes and the imported `deprecated` function
    for item in items.keys() | [deprecated.__name__]:
        if item in mod_globals:
            del mod_globals[item]


__all__ = [
    "BaseExceptionGroup",
    "ExceptionGroup",
    "deprecated",
    "detect_encoding",
    "is_linux",
    "is_android",
    "is_darwin",
    "is_freebsd",
    "is_posix",
    "is_win32",
]



"""
$description Chinese live-streaming platform for live video game broadcasts and individual live streams.
$url huajiao.com
$type live
$metadata author
$metadata category
$metadata title
"""

import base64
import random
import re
import time
import uuid

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?huajiao\.com/l/(?P<channel>[^/]+)"),
)
class Huajiao(Plugin):
    URL_LAPI = "https://g2.live.360.cn/liveplay"

    def _get_streams(self):
        data = self.session.http.get(
            self.url,
            schema=validate.Schema(
                re.compile(r"var\s*feed\s*=\s*(?P<feed>{.+?})\s*;", re.DOTALL),
                validate.none_or_all(
                    validate.get("feed"),
                    validate.parse_json(),
                    {
                        "author": {
                            "nickname": str,
                        },
                        "feed": {
                            "title": str,
                            "game": str,
                            "m3u8": validate.any("", validate.url()),
                            "sn": str,
                        },
                        "relay": {
                            "channel": str,
                        },
                    },
                    validate.union_get(
                        ("author", "nickname"),
                        ("feed", "title"),
                        ("feed", "game"),
                        ("feed", "m3u8"),
                        ("feed", "sn"),
                        ("relay", "channel"),
                    ),
                ),
            ),
        )
        if not data:
            return

        self.author, self.title, self.category, m3u8, sn, channel_sid = data

        if m3u8:
            return HLSStream(self.session, m3u8)

        stream_url = self.session.http.get(
            self.URL_LAPI,
            params={
                "stype": "flv",
                "channel": channel_sid,
                "bid": "huajiao",
                "sn": sn,
                "sid": uuid.uuid4().hex.upper(),
                "_rate": "xd",
                "ts": time.time(),
                "r": random.random(),
            },
            schema=validate.Schema(
                validate.transform(lambda text: base64.b64decode(text[0:3] + text[6:]).decode("utf-8")),
                validate.parse_json(),
                {"main": validate.url()},
                validate.get("main"),
            ),
        )
        return {"live": HTTPStream(self.session, stream_url)}


__plugin__ = Huajiao



from dataclasses import dataclass, field
from urllib.parse import urlparse

from streamlink.utils.dataclass import FormattedDataclass


@dataclass(kw_only=True)
class Segment(metaclass=FormattedDataclass, extra=["fileext"]):
    num: int
    init: bool = False
    discontinuity: bool = False
    uri: str = field(repr=False)
    duration: float

    @property
    def fileext(self) -> str | None:
        path = urlparse(self.uri).path
        ext = path.split(".")[-1]
        if 4 >= len(ext) >= 2:
            return ext
        return None



"""
$description Chinese live-streaming platform for live video game broadcasts and individual live streams.
$url huya.com
$type live
$metadata id
$metadata author
$metadata title
"""

from __future__ import annotations

import base64
import hashlib
import random
import re
import sys
import time
from html import unescape as html_unescape
from urllib.parse import parse_qsl, unquote

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.http import HTTPStream
from streamlink.utils.url import update_scheme


log = getLogger(__name__)


@pluginmatcher(
    re.compile(
        r"https?://(?:www\.)?huya\.com/(?P<channel>[^/?]+)",
    ),
)
class Huya(Plugin):
    QUALITY_WEIGHTS: dict[str, int] = {}

    _STREAM_URL_QUERYSTRING_PARAMS = "wsTime", "fm", "ctype", "fs"

    _CONSTANTS = {
        "t": 100,
        "ver": 1,
        "sv": 2401090219,
        "codec": 264,
    }

    @classmethod
    def stream_weight(cls, stream: str) -> tuple[float, str]:
        weight = cls.QUALITY_WEIGHTS.get(stream)
        if weight:
            return weight, "huya"

        return super().stream_weight(stream)

    def _get_streams(self):
        data = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//script[contains(text(),'var hyPlayerConfig = {')][1]/text()"),
                validate.none_or_all(
                    re.compile(r"""(?P<q>"?)stream(?P=q)\s*:\s*(?:"(?P<base64>.+?)"|(?P<json>\{.+?})\s*}\s*;)"""),
                ),
                validate.none_or_all(
                    validate.any(
                        validate.all(
                            validate.get("base64"),
                            str,
                            validate.transform(base64.b64decode),
                        ),
                        validate.all(
                            validate.get("json"),
                            str,
                        ),
                    ),
                    validate.parse_json(),
                    {
                        "data": [
                            {
                                "gameLiveInfo": {
                                    "liveId": str,
                                    "nick": str,
                                    "roomName": str,
                                },
                                "gameStreamInfoList": [
                                    validate.all(
                                        {
                                            "sCdnType": str,
                                            "sStreamName": str,
                                            "sFlvUrl": str,
                                            "sFlvUrlSuffix": str,
                                            "sFlvAntiCode": validate.all(str, validate.transform(html_unescape)),
                                        },
                                        validate.union_get(
                                            "sCdnType",
                                            "sStreamName",
                                            "sFlvUrl",
                                            "sFlvUrlSuffix",
                                            "sFlvAntiCode",
                                        ),
                                    ),
                                ],
                            },
                        ],
                        "vMultiStreamInfo": [{"iBitRate": int}],
                    },
                    validate.union_get(
                        ("data", 0, "gameLiveInfo", "liveId"),
                        ("data", 0, "gameLiveInfo", "nick"),
                        ("data", 0, "gameLiveInfo", "roomName"),
                        ("data", 0, "gameStreamInfoList"),
                        "vMultiStreamInfo",
                    ),
                ),
            ),
        )
        if not data:
            return

        self.id, self.author, self.title, streamdata, v_multi_stream_info = data

        self.session.http.headers.update({
            "Origin": "https://www.huya.com",
            "Referer": "https://www.huya.com/",
        })

        for cdn_type, stream_name, flvurl, suffix, anticode in streamdata:
            for v_stream_info in v_multi_stream_info:
                i_bit_rate = v_stream_info["iBitRate"]
                qs = {k: v for k, v in dict(parse_qsl(anticode)).items() if k in self._STREAM_URL_QUERYSTRING_PARAMS}
                url = update_scheme("https://", f"{flvurl}/{stream_name}.{suffix}")
                params = self._get_stream_params(
                    qs.get("fm", ""),
                    qs.get("fs", ""),
                    qs.get("ctype", "huya_live"),
                    qs.get("wsTime", ""),
                    stream_name,
                    i_bit_rate,
                )
                name = f"{cdn_type.lower()}_{'source' if i_bit_rate == 0 else f'{i_bit_rate}k'}"
                weight = sys.maxsize if i_bit_rate == 0 else i_bit_rate
                self.QUALITY_WEIGHTS[name] = weight
                yield name, HTTPStream(self.session, url, params=params)

        log.debug("QUALITY_WEIGHTS: %r", self.QUALITY_WEIGHTS)

    def _get_stream_params(self, fm, fs, ctype, ws_time, stream_name, i_bit_rate):
        uid = random.randint(12340000, 12349999)
        convert_uid = (uid << 8 | uid >> (32 - 8)) & 0xFFFFFFFF
        timestamp = int(time.time() * 1000)
        seqid = uid + timestamp
        ws_secret_prefix = base64.b64decode(unquote(fm).encode()).decode().split("_")[0]
        ws_secret_hash = hashlib.md5(f"{seqid}|{ctype}|{self._CONSTANTS['t']}".encode()).hexdigest()
        ws_secret = hashlib.md5(
            f"{ws_secret_prefix}_{convert_uid}_{stream_name}_{ws_secret_hash}_{ws_time}".encode(),
        ).hexdigest()
        params = {
            "wsSecret": ws_secret,
            "wsTime": ws_time,
            "ctype": ctype,
            "fs": fs,
            "seqid": seqid,
            "u": convert_uid,
            "sdk_sid": timestamp,
            "ratio": i_bit_rate,
        }
        params.update(self._CONSTANTS)
        return params


__plugin__ = Huya



"""
$description German sports magazine live streams and VOD content, owned by ARD.
$url sportschau.de
$type live
$metadata id
$metadata title
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream


log = getLogger(__name__)


@pluginmatcher(re.compile(r"https?://(?:\w+\.)*sportschau\.de/"))
class Sportschau(Plugin):
    _schema_media = validate.Schema(
        validate.parse_html(),
        validate.xml_xpath_string(".//*[@data-v-type='MediaPlayer'][@data-v][1]/@data-v"),
        validate.none_or_all(
            validate.parse_json(),
            {
                "mc": {
                    validate.optional("id"): str,
                    "meta": {
                        "title": str,
                    },
                    "streams": [
                        validate.all(
                            {
                                "media": [
                                    validate.all(
                                        {
                                            "mimeType": str,
                                            "url": validate.url(),
                                        },
                                        validate.union_get(
                                            "mimeType",
                                            "url",
                                        ),
                                    ),
                                ],
                            },
                            validate.get("media"),
                        ),
                    ],
                },
            },
            validate.get("mc"),
            validate.union_get(
                "id",
                ("meta", "title"),
                "streams",
            ),
        ),
    )

    def _get_streams(self):
        data = self.session.http.get(self.url, schema=self._schema_media)
        if not data:
            return

        self.id, self.title, streams = data

        for media in streams:
            for mime_type, url in media:
                if mime_type == "application/vnd.apple.mpegurl":
                    yield from HLSStream.parse_variant_playlist(self.session, url).items()
                elif mime_type.startswith("audio/"):
                    yield "audio", HTTPStream(self.session, url)


__plugin__ = Sportschau



# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all modules.
#
# CDP version: v0.0.1510116
# CDP domain: Browser

from __future__ import annotations

import enum
from collections.abc import Generator
from dataclasses import dataclass
from typing import Any

import streamlink.webbrowser.cdp.devtools.page as page
import streamlink.webbrowser.cdp.devtools.target as target
from streamlink.webbrowser.cdp.devtools.util import T_JSON_DICT, CDPEvent


class BrowserContextID(str):
    def to_json(self) -> str:
        return self

    @classmethod
    def from_json(cls, json: str) -> BrowserContextID:
        return cls(json)

    def __repr__(self):
        return f"BrowserContextID({super().__repr__()})"


class WindowID(int):
    def to_json(self) -> int:
        return self

    @classmethod
    def from_json(cls, json: int) -> WindowID:
        return cls(json)

    def __repr__(self):
        return f"WindowID({super().__repr__()})"


class WindowState(enum.Enum):
    """
    The state of the browser window.
    """
    NORMAL = "normal"
    MINIMIZED = "minimized"
    MAXIMIZED = "maximized"
    FULLSCREEN = "fullscreen"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> WindowState:
        return cls(json)


@dataclass
class Bounds:
    """
    Browser window bounds information
    """
    #: The offset from the left edge of the screen to the window in pixels.
    left: int | None = None

    #: The offset from the top edge of the screen to the window in pixels.
    top: int | None = None

    #: The window width in pixels.
    width: int | None = None

    #: The window height in pixels.
    height: int | None = None

    #: The window state. Default to normal.
    window_state: WindowState | None = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        if self.left is not None:
            json["left"] = self.left
        if self.top is not None:
            json["top"] = self.top
        if self.width is not None:
            json["width"] = self.width
        if self.height is not None:
            json["height"] = self.height
        if self.window_state is not None:
            json["windowState"] = self.window_state.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> Bounds:
        return cls(
            left=int(json["left"]) if "left" in json else None,
            top=int(json["top"]) if "top" in json else None,
            width=int(json["width"]) if "width" in json else None,
            height=int(json["height"]) if "height" in json else None,
            window_state=WindowState.from_json(json["windowState"]) if "windowState" in json else None,
        )


class PermissionType(enum.Enum):
    AR = "ar"
    AUDIO_CAPTURE = "audioCapture"
    AUTOMATIC_FULLSCREEN = "automaticFullscreen"
    BACKGROUND_FETCH = "backgroundFetch"
    BACKGROUND_SYNC = "backgroundSync"
    CAMERA_PAN_TILT_ZOOM = "cameraPanTiltZoom"
    CAPTURED_SURFACE_CONTROL = "capturedSurfaceControl"
    CLIPBOARD_READ_WRITE = "clipboardReadWrite"
    CLIPBOARD_SANITIZED_WRITE = "clipboardSanitizedWrite"
    DISPLAY_CAPTURE = "displayCapture"
    DURABLE_STORAGE = "durableStorage"
    GEOLOCATION = "geolocation"
    HAND_TRACKING = "handTracking"
    IDLE_DETECTION = "idleDetection"
    KEYBOARD_LOCK = "keyboardLock"
    LOCAL_FONTS = "localFonts"
    LOCAL_NETWORK_ACCESS = "localNetworkAccess"
    MIDI = "midi"
    MIDI_SYSEX = "midiSysex"
    NFC = "nfc"
    NOTIFICATIONS = "notifications"
    PAYMENT_HANDLER = "paymentHandler"
    PERIODIC_BACKGROUND_SYNC = "periodicBackgroundSync"
    POINTER_LOCK = "pointerLock"
    PROTECTED_MEDIA_IDENTIFIER = "protectedMediaIdentifier"
    SENSORS = "sensors"
    SMART_CARD = "smartCard"
    SPEAKER_SELECTION = "speakerSelection"
    STORAGE_ACCESS = "storageAccess"
    TOP_LEVEL_STORAGE_ACCESS = "topLevelStorageAccess"
    VIDEO_CAPTURE = "videoCapture"
    VR = "vr"
    WAKE_LOCK_SCREEN = "wakeLockScreen"
    WAKE_LOCK_SYSTEM = "wakeLockSystem"
    WEB_APP_INSTALLATION = "webAppInstallation"
    WEB_PRINTING = "webPrinting"
    WINDOW_MANAGEMENT = "windowManagement"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> PermissionType:
        return cls(json)


class PermissionSetting(enum.Enum):
    GRANTED = "granted"
    DENIED = "denied"
    PROMPT = "prompt"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> PermissionSetting:
        return cls(json)


@dataclass
class PermissionDescriptor:
    """
    Definition of PermissionDescriptor defined in the Permissions API:
    https://w3c.github.io/permissions/#dom-permissiondescriptor.
    """
    #: Name of permission.
    #: See https://cs.chromium.org/chromium/src/third_party/blink/renderer/modules/permissions/permission_descriptor.idl for valid permission names.
    name: str

    #: For "midi" permission, may also specify sysex control.
    sysex: bool | None = None

    #: For "push" permission, may specify userVisibleOnly.
    #: Note that userVisibleOnly = true is the only currently supported type.
    user_visible_only: bool | None = None

    #: For "clipboard" permission, may specify allowWithoutSanitization.
    allow_without_sanitization: bool | None = None

    #: For "fullscreen" permission, must specify allowWithoutGesture:true.
    allow_without_gesture: bool | None = None

    #: For "camera" permission, may specify panTiltZoom.
    pan_tilt_zoom: bool | None = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["name"] = self.name
        if self.sysex is not None:
            json["sysex"] = self.sysex
        if self.user_visible_only is not None:
            json["userVisibleOnly"] = self.user_visible_only
        if self.allow_without_sanitization is not None:
            json["allowWithoutSanitization"] = self.allow_without_sanitization
        if self.allow_without_gesture is not None:
            json["allowWithoutGesture"] = self.allow_without_gesture
        if self.pan_tilt_zoom is not None:
            json["panTiltZoom"] = self.pan_tilt_zoom
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> PermissionDescriptor:
        return cls(
            name=str(json["name"]),
            sysex=bool(json["sysex"]) if "sysex" in json else None,
            user_visible_only=bool(json["userVisibleOnly"]) if "userVisibleOnly" in json else None,
            allow_without_sanitization=bool(json["allowWithoutSanitization"]) if "allowWithoutSanitization" in json else None,
            allow_without_gesture=bool(json["allowWithoutGesture"]) if "allowWithoutGesture" in json else None,
            pan_tilt_zoom=bool(json["panTiltZoom"]) if "panTiltZoom" in json else None,
        )


class BrowserCommandId(enum.Enum):
    """
    Browser command ids used by executeBrowserCommand.
    """
    OPEN_TAB_SEARCH = "openTabSearch"
    CLOSE_TAB_SEARCH = "closeTabSearch"
    OPEN_GLIC = "openGlic"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> BrowserCommandId:
        return cls(json)


@dataclass
class Bucket:
    """
    Chrome histogram bucket.
    """
    #: Minimum value (inclusive).
    low: int

    #: Maximum value (exclusive).
    high: int

    #: Number of samples.
    count: int

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["low"] = self.low
        json["high"] = self.high
        json["count"] = self.count
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> Bucket:
        return cls(
            low=int(json["low"]),
            high=int(json["high"]),
            count=int(json["count"]),
        )


@dataclass
class Histogram:
    """
    Chrome histogram.
    """
    #: Name.
    name: str

    #: Sum of sample values.
    sum_: int

    #: Total number of samples.
    count: int

    #: Buckets.
    buckets: list[Bucket]

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["name"] = self.name
        json["sum"] = self.sum_
        json["count"] = self.count
        json["buckets"] = [i.to_json() for i in self.buckets]
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> Histogram:
        return cls(
            name=str(json["name"]),
            sum_=int(json["sum"]),
            count=int(json["count"]),
            buckets=[Bucket.from_json(i) for i in json["buckets"]],
        )


class PrivacySandboxAPI(enum.Enum):
    BIDDING_AND_AUCTION_SERVICES = "BiddingAndAuctionServices"
    TRUSTED_KEY_VALUE = "TrustedKeyValue"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> PrivacySandboxAPI:
        return cls(json)


def set_permission(
    permission: PermissionDescriptor,
    setting: PermissionSetting,
    origin: str | None = None,
    embedding_origin: str | None = None,
    browser_context_id: BrowserContextID | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Set permission settings for given requesting and embedding origins.

    **EXPERIMENTAL**

    :param permission: Descriptor of permission to override.
    :param setting: Setting of the permission.
    :param origin: *(Optional)* Requesting origin the permission applies to, all origins if not specified.
    :param embedding_origin: *(Optional)* Embedding origin the permission applies to. It is ignored unless the requesting origin is present and valid. If the requesting origin is provided but the embedding origin isn't, the requesting origin is used as the embedding origin.
    :param browser_context_id: *(Optional)* Context to override. When omitted, default browser context is used.
    """
    params: T_JSON_DICT = {}
    params["permission"] = permission.to_json()
    params["setting"] = setting.to_json()
    if origin is not None:
        params["origin"] = origin
    if embedding_origin is not None:
        params["embeddingOrigin"] = embedding_origin
    if browser_context_id is not None:
        params["browserContextId"] = browser_context_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.setPermission",
        "params": params,
    }
    yield cmd_dict


def grant_permissions(
    permissions: list[PermissionType],
    origin: str | None = None,
    browser_context_id: BrowserContextID | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Grant specific permissions to the given origin and reject all others.

    **EXPERIMENTAL**

    :param permissions:
    :param origin: *(Optional)* Origin the permission applies to, all origins if not specified.
    :param browser_context_id: *(Optional)* BrowserContext to override permissions. When omitted, default browser context is used.
    """
    params: T_JSON_DICT = {}
    params["permissions"] = [i.to_json() for i in permissions]
    if origin is not None:
        params["origin"] = origin
    if browser_context_id is not None:
        params["browserContextId"] = browser_context_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.grantPermissions",
        "params": params,
    }
    yield cmd_dict


def reset_permissions(
    browser_context_id: BrowserContextID | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Reset all permission management for all origins.

    :param browser_context_id: *(Optional)* BrowserContext to reset permissions. When omitted, default browser context is used.
    """
    params: T_JSON_DICT = {}
    if browser_context_id is not None:
        params["browserContextId"] = browser_context_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.resetPermissions",
        "params": params,
    }
    yield cmd_dict


def set_download_behavior(
    behavior: str,
    browser_context_id: BrowserContextID | None = None,
    download_path: str | None = None,
    events_enabled: bool | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Set the behavior when downloading a file.

    **EXPERIMENTAL**

    :param behavior: Whether to allow all or deny all download requests, or use default Chrome behavior if available (otherwise deny). ``allowAndName`` allows download and names files according to their download guids.
    :param browser_context_id: *(Optional)* BrowserContext to set download behavior. When omitted, default browser context is used.
    :param download_path: *(Optional)* The default path to save downloaded files to. This is required if behavior is set to 'allow' or 'allowAndName'.
    :param events_enabled: *(Optional)* Whether to emit download events (defaults to false).
    """
    params: T_JSON_DICT = {}
    params["behavior"] = behavior
    if browser_context_id is not None:
        params["browserContextId"] = browser_context_id.to_json()
    if download_path is not None:
        params["downloadPath"] = download_path
    if events_enabled is not None:
        params["eventsEnabled"] = events_enabled
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.setDownloadBehavior",
        "params": params,
    }
    yield cmd_dict


def cancel_download(
    guid: str,
    browser_context_id: BrowserContextID | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Cancel a download if in progress

    **EXPERIMENTAL**

    :param guid: Global unique identifier of the download.
    :param browser_context_id: *(Optional)* BrowserContext to perform the action in. When omitted, default browser context is used.
    """
    params: T_JSON_DICT = {}
    params["guid"] = guid
    if browser_context_id is not None:
        params["browserContextId"] = browser_context_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.cancelDownload",
        "params": params,
    }
    yield cmd_dict


def close() -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Close browser gracefully.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.close",
    }
    yield cmd_dict


def crash() -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Crashes browser on the main thread.

    **EXPERIMENTAL**
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.crash",
    }
    yield cmd_dict


def crash_gpu_process() -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Crashes GPU process.

    **EXPERIMENTAL**
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.crashGpuProcess",
    }
    yield cmd_dict


def get_version() -> Generator[T_JSON_DICT, T_JSON_DICT, tuple[str, str, str, str, str]]:
    """
    Returns version information.

    :returns: A tuple with the following items:

        0. **protocolVersion** - Protocol version.
        1. **product** - Product name.
        2. **revision** - Product revision.
        3. **userAgent** - User-Agent.
        4. **jsVersion** - V8 version.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.getVersion",
    }
    json = yield cmd_dict
    return (
        str(json["protocolVersion"]),
        str(json["product"]),
        str(json["revision"]),
        str(json["userAgent"]),
        str(json["jsVersion"]),
    )


def get_browser_command_line() -> Generator[T_JSON_DICT, T_JSON_DICT, list[str]]:
    """
    Returns the command line switches for the browser process if, and only if
    --enable-automation is on the commandline.

    **EXPERIMENTAL**

    :returns: Commandline parameters
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.getBrowserCommandLine",
    }
    json = yield cmd_dict
    return [str(i) for i in json["arguments"]]


def get_histograms(
    query: str | None = None,
    delta: bool | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, list[Histogram]]:
    """
    Get Chrome histograms.

    **EXPERIMENTAL**

    :param query: *(Optional)* Requested substring in name. Only histograms which have query as a substring in their name are extracted. An empty or absent query returns all histograms.
    :param delta: *(Optional)* If true, retrieve delta since last delta call.
    :returns: Histograms.
    """
    params: T_JSON_DICT = {}
    if query is not None:
        params["query"] = query
    if delta is not None:
        params["delta"] = delta
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.getHistograms",
        "params": params,
    }
    json = yield cmd_dict
    return [Histogram.from_json(i) for i in json["histograms"]]


def get_histogram(
    name: str,
    delta: bool | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, Histogram]:
    """
    Get a Chrome histogram by name.

    **EXPERIMENTAL**

    :param name: Requested histogram name.
    :param delta: *(Optional)* If true, retrieve delta since last delta call.
    :returns: Histogram.
    """
    params: T_JSON_DICT = {}
    params["name"] = name
    if delta is not None:
        params["delta"] = delta
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.getHistogram",
        "params": params,
    }
    json = yield cmd_dict
    return Histogram.from_json(json["histogram"])


def get_window_bounds(
    window_id: WindowID,
) -> Generator[T_JSON_DICT, T_JSON_DICT, Bounds]:
    """
    Get position and size of the browser window.

    **EXPERIMENTAL**

    :param window_id: Browser window id.
    :returns: Bounds information of the window. When window state is 'minimized', the restored window position and size are returned.
    """
    params: T_JSON_DICT = {}
    params["windowId"] = window_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.getWindowBounds",
        "params": params,
    }
    json = yield cmd_dict
    return Bounds.from_json(json["bounds"])


def get_window_for_target(
    target_id: target.TargetID | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, tuple[WindowID, Bounds]]:
    """
    Get the browser window that contains the devtools target.

    **EXPERIMENTAL**

    :param target_id: *(Optional)* Devtools agent host id. If called as a part of the session, associated targetId is used.
    :returns: A tuple with the following items:

        0. **windowId** - Browser window id.
        1. **bounds** - Bounds information of the window. When window state is 'minimized', the restored window position and size are returned.
    """
    params: T_JSON_DICT = {}
    if target_id is not None:
        params["targetId"] = target_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.getWindowForTarget",
        "params": params,
    }
    json = yield cmd_dict
    return (
        WindowID.from_json(json["windowId"]),
        Bounds.from_json(json["bounds"]),
    )


def set_window_bounds(
    window_id: WindowID,
    bounds: Bounds,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Set position and/or size of the browser window.

    **EXPERIMENTAL**

    :param window_id: Browser window id.
    :param bounds: New window bounds. The 'minimized', 'maximized' and 'fullscreen' states cannot be combined with 'left', 'top', 'width' or 'height'. Leaves unspecified fields unchanged.
    """
    params: T_JSON_DICT = {}
    params["windowId"] = window_id.to_json()
    params["bounds"] = bounds.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.setWindowBounds",
        "params": params,
    }
    yield cmd_dict


def set_contents_size(
    window_id: WindowID,
    width: int | None = None,
    height: int | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Set size of the browser contents resizing browser window as necessary.

    **EXPERIMENTAL**

    :param window_id: Browser window id.
    :param width: *(Optional)* The window contents width in DIP. Assumes current width if omitted. Must be specified if 'height' is omitted.
    :param height: *(Optional)* The window contents height in DIP. Assumes current height if omitted. Must be specified if 'width' is omitted.
    """
    params: T_JSON_DICT = {}
    params["windowId"] = window_id.to_json()
    if width is not None:
        params["width"] = width
    if height is not None:
        params["height"] = height
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.setContentsSize",
        "params": params,
    }
    yield cmd_dict


def set_dock_tile(
    badge_label: str | None = None,
    image: str | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Set dock tile details, platform-specific.

    **EXPERIMENTAL**

    :param badge_label: *(Optional)*
    :param image: *(Optional)* Png encoded image. (Encoded as a base64 string when passed over JSON)
    """
    params: T_JSON_DICT = {}
    if badge_label is not None:
        params["badgeLabel"] = badge_label
    if image is not None:
        params["image"] = image
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.setDockTile",
        "params": params,
    }
    yield cmd_dict


def execute_browser_command(
    command_id: BrowserCommandId,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Invoke custom browser commands used by telemetry.

    **EXPERIMENTAL**

    :param command_id:
    """
    params: T_JSON_DICT = {}
    params["commandId"] = command_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.executeBrowserCommand",
        "params": params,
    }
    yield cmd_dict


def add_privacy_sandbox_enrollment_override(
    url: str,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Allows a site to use privacy sandbox features that require enrollment
    without the site actually being enrolled. Only supported on page targets.

    :param url:
    """
    params: T_JSON_DICT = {}
    params["url"] = url
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.addPrivacySandboxEnrollmentOverride",
        "params": params,
    }
    yield cmd_dict


def add_privacy_sandbox_coordinator_key_config(
    api: PrivacySandboxAPI,
    coordinator_origin: str,
    key_config: str,
    browser_context_id: BrowserContextID | None = None,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Configures encryption keys used with a given privacy sandbox API to talk
    to a trusted coordinator.  Since this is intended for test automation only,
    coordinatorOrigin must be a .test domain. No existing coordinator
    configuration for the origin may exist.

    :param api:
    :param coordinator_origin:
    :param key_config:
    :param browser_context_id: *(Optional)* BrowserContext to perform the action in. When omitted, default browser context is used.
    """
    params: T_JSON_DICT = {}
    params["api"] = api.to_json()
    params["coordinatorOrigin"] = coordinator_origin
    params["keyConfig"] = key_config
    if browser_context_id is not None:
        params["browserContextId"] = browser_context_id.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Browser.addPrivacySandboxCoordinatorKeyConfig",
        "params": params,
    }
    yield cmd_dict


@dataclass
class DownloadWillBegin(CDPEvent, event="Browser.downloadWillBegin"):
    """
    **EXPERIMENTAL**

    Fired when page is about to start a download.
    """
    #: Id of the frame that caused the download to begin.
    frame_id: page.FrameId
    #: Global unique identifier of the download.
    guid: str
    #: URL of the resource being downloaded.
    url: str
    #: Suggested file name of the resource (the actual name of the file saved on disk may differ).
    suggested_filename: str

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> DownloadWillBegin:
        return cls(
            frame_id=page.FrameId.from_json(json["frameId"]),
            guid=str(json["guid"]),
            url=str(json["url"]),
            suggested_filename=str(json["suggestedFilename"]),
        )


@dataclass
class DownloadProgress(CDPEvent, event="Browser.downloadProgress"):
    """
    **EXPERIMENTAL**

    Fired when download makes progress. Last call has ``done`` == true.
    """
    #: Global unique identifier of the download.
    guid: str
    #: Total expected bytes to download.
    total_bytes: float
    #: Total bytes received.
    received_bytes: float
    #: Download status.
    state: str
    #: If download is "completed", provides the path of the downloaded file.
    #: Depending on the platform, it is not guaranteed to be set, nor the file
    #: is guaranteed to exist.
    file_path: str | None

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> DownloadProgress:
        return cls(
            guid=str(json["guid"]),
            total_bytes=float(json["totalBytes"]),
            received_bytes=float(json["receivedBytes"]),
            state=str(json["state"]),
            file_path=str(json["filePath"]) if "filePath" in json else None,
        )



"""
$description A network of live webcams for tourism and entertainment.
$url earthcam.com
$type live, vod
$metadata author
$metadata category
$metadata title
$notes Only works for the cams hosted on EarthCam
"""

import re
from urllib.parse import urlparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.parse import parse_qsd
from streamlink.utils.url import update_scheme


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?earthcam\.com/"),
)
class EarthCam(Plugin):
    def _get_streams(self):
        data = self.session.http.get(
            self.url,
            schema=validate.Schema(
                re.compile(r"""var\s+json_base\s*=\s*(?P<json>{.*?});""", re.DOTALL),
                validate.none_or_all(
                    validate.get("json"),
                    validate.parse_json(),
                    {
                        "cam": {
                            str: {
                                "live_type": str,
                                "html5_streamingdomain": str,
                                "html5_streampath": str,
                                "group": str,
                                "location": str,
                                "title": str,
                                "liveon": str,
                                "defaulttab": str,
                            },
                        },
                    },
                    validate.get("cam"),
                ),
            ),
        )
        if not data:
            return

        cam_name = parse_qsd(urlparse(self.url).query).get("cam") or next(iter(data.keys()), None)
        cam_data = data.get(cam_name)
        if not cam_data:
            return

        # exclude everything other than live video streams
        if cam_data["live_type"] != "flashvideo" or cam_data["liveon"] != "true" or cam_data["defaulttab"] != "live":
            return

        log.debug(f"Found cam {cam_name}")
        hls_domain = cam_data["html5_streamingdomain"]
        hls_playpath = cam_data["html5_streampath"]

        self.author = cam_data["group"]
        self.category = cam_data["location"]
        self.title = cam_data["title"]

        if hls_playpath:
            hls_url = update_scheme("https://", f"{hls_domain}{hls_playpath}")
            yield from HLSStream.parse_variant_playlist(self.session, hls_url).items()


__plugin__ = EarthCam



"""
$description Live TV channels and video on-demand service from NOS, a Dutch public, state-owned broadcaster.
$url nos.nl
$type live, vod
$metadata id
$metadata title
$region Netherlands
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:\w+\.)?nos\.nl/(?:live$|livestream/|l/|video/|collectie/)"),
)
class NOS(Plugin):
    def _get_streams(self):
        data = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//script[@type='application/ld+json'][1]/text()"),
                validate.none_or_all(
                    validate.parse_json(),
                    {
                        "@type": validate.any(
                            "VideoObject",
                            validate.all(
                                list,
                                validate.contains("VideoObject"),
                            ),
                        ),
                        "encodingFormat": "application/vnd.apple.mpegurl",
                        "contentUrl": validate.url(),
                        "identifier": validate.any(int, str),
                        "name": str,
                    },
                    validate.union_get(
                        "contentUrl",
                        "identifier",
                        "name",
                    ),
                ),
            ),
        )
        if not data:
            return

        hls_url, self.id, self.title = data

        res = self.session.http.get(hls_url, raise_for_status=False)
        if res.status_code >= 400:
            log.error("Content is inaccessible or may have expired")
            return

        return HLSStream.parse_variant_playlist(self.session, hls_url)


__plugin__ = NOS



from streamlink.stream.dash.dash import DASHStream, DASHStreamReader, DASHStreamWorker, DASHStreamWriter
from streamlink.stream.dash.manifest import MPD, MPDParsingError
from streamlink.stream.dash.segment import DASHSegment



"""
$description United Arab Emirates CDN hosting live content for various websites in The Middle East.
$url alwasat.ly
$type live
$region various
"""

import base64
import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(name="alwasatly", pattern=re.compile(r"https?://(?:www\.)?alwasat\.ly"))
class HiPlayer(Plugin):
    DAI_URL = "https://pubads.g.doubleclick.net/ssai/event/{0}/streams"

    def _get_streams(self):
        js_url = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//script[contains(text(), 'https://hiplayer.hibridcdn.net/l/')]/text()"),
                validate.none_or_all(
                    re.compile(r"""(?P<q>['"])(?P<url>https://hiplayer.hibridcdn.net/l/.+?)(?P=q)"""),
                    validate.none_or_all(
                        validate.get("url"),
                        validate.url(),
                    ),
                ),
            ),
        )
        if not js_url:
            log.error("Could not find JS URL")
            return

        log.debug(f"JS URL={js_url}")

        data = self.session.http.get(
            js_url,
            schema=validate.Schema(
                re.compile(r"\[(?P<data>[^]]+)]\.join\([\"']{2}\)"),
                validate.none_or_all(
                    validate.get("data"),
                    validate.transform(lambda s: re.sub(r"['\", ]", "", s)),
                    validate.transform(base64.b64decode),
                    validate.parse_json(),
                    validate.any(
                        None,
                        {
                            "daiEnabled": bool,
                            "daiAssetKey": str,
                            "daiApiKey": str,
                            "streamUrl": validate.any(validate.url(), ""),
                        },
                    ),
                ),
            ),
        )
        if not data:
            log.error("Could not find base64 encoded JSON data")
            return

        hls_url = data["streamUrl"]

        if data["daiEnabled"]:
            log.debug("daiEnabled=true")
            hls_url = self.session.http.post(
                self.DAI_URL.format(data["daiAssetKey"]),
                data={"api-key": data["daiApiKey"]},
                schema=validate.Schema(
                    validate.parse_json(),
                    {
                        "stream_manifest": validate.url(),
                    },
                    validate.get("stream_manifest"),
                ),
            )

        if hls_url:
            return HLSStream.parse_variant_playlist(self.session, hls_url)


__plugin__ = HiPlayer



from __future__ import annotations

import os
import tempfile
from pathlib import Path

from streamlink.compat import is_darwin, is_win32


DEFAULT_STREAM_METADATA = {
    "id": "Unknown ID",
    "title": "Unknown Title",
    "author": "Unknown Author",
    "category": "No Category",
    "game": "No Game/Category",
}

PROGRESS_INTERVAL_NO_STATUS = 2

CONFIG_FILES: list[Path]
PLUGIN_DIRS: list[Path]
LOG_DIR: Path

if is_win32:
    APPDATA = Path(os.environ.get("APPDATA") or Path.home() / "AppData")
    CONFIG_FILES = [
        APPDATA / "streamlink" / "config",
    ]
    PLUGIN_DIRS = [
        APPDATA / "streamlink" / "plugins",
    ]
    LOG_DIR = Path(tempfile.gettempdir()) / "streamlink" / "logs"
elif is_darwin:
    XDG_CONFIG_HOME = Path(os.environ.get("XDG_CONFIG_HOME", "~/.config")).expanduser()
    CONFIG_FILES = [
        Path.home() / "Library" / "Application Support" / "streamlink" / "config",
    ]
    PLUGIN_DIRS = [
        Path.home() / "Library" / "Application Support" / "streamlink" / "plugins",
    ]
    LOG_DIR = Path.home() / "Library" / "Logs" / "streamlink"
else:
    XDG_CONFIG_HOME = Path(os.environ.get("XDG_CONFIG_HOME", "~/.config")).expanduser()
    XDG_DATA_HOME = Path(os.environ.get("XDG_DATA_HOME", "~/.local/share")).expanduser()
    XDG_STATE_HOME = Path(os.environ.get("XDG_STATE_HOME", "~/.local/state")).expanduser()
    CONFIG_FILES = [
        XDG_CONFIG_HOME / "streamlink" / "config",
    ]
    PLUGIN_DIRS = [
        XDG_DATA_HOME / "streamlink" / "plugins",
    ]
    LOG_DIR = XDG_STATE_HOME / "streamlink" / "logs"

STREAM_SYNONYMS = ["best", "worst", "best-unfiltered", "worst-unfiltered"]
STREAM_PASSTHROUGH = ["hls", "http"]


__all__ = [
    "CONFIG_FILES",
    "DEFAULT_STREAM_METADATA",
    "LOG_DIR",
    "PLUGIN_DIRS",
    "PROGRESS_INTERVAL_NO_STATUS",
    "STREAM_PASSTHROUGH",
    "STREAM_SYNONYMS",
]



import sys
from os import devnull
from typing import BinaryIO


try:
    stdout: BinaryIO = sys.stdout.buffer
except AttributeError:  # pragma: no cover
    from atexit import register as _atexit_register
    from pathlib import Path

    stdout = Path(devnull).open("wb")
    _atexit_register(stdout.close)
    del _atexit_register


__all__ = [
    "stdout",
]



from __future__ import annotations

import os
from contextlib import asynccontextmanager
from pathlib import Path
from typing import TYPE_CHECKING

import trio

import streamlink.validate as validate
from streamlink.compat import is_darwin, is_win32
from streamlink.utils.socket import find_free_port_ipv4, find_free_port_ipv6
from streamlink.webbrowser.webbrowser import Webbrowser


if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

    from streamlink.session import Streamlink


class ChromiumWebbrowser(Webbrowser):
    ERROR_RESOLVE = "Could not find Chromium-based web browser executable"

    @classmethod
    def names(cls) -> list[str]:
        return [
            "chromium",
            "chromium-browser",
            "chrome",
            "google-chrome",
            "google-chrome-stable",
        ]

    @classmethod
    def fallback_paths(cls) -> list[str | Path]:
        if is_win32:
            ms_edge: list[str | Path] = [
                str(Path(base) / sub / "msedge.exe")
                for sub in (
                    "Microsoft\\Edge\\Application",
                    "Microsoft\\Edge Beta\\Application",
                    "Microsoft\\Edge Dev\\Application",
                )
                for base in [
                    os.getenv(env)
                    for env in (
                        "PROGRAMFILES",
                        "PROGRAMFILES(X86)",
                    )
                ]
                if base is not None
            ]
            google_chrome: list[str | Path] = [
                str(Path(base) / sub / "chrome.exe")
                for sub in (
                    "Google\\Chrome\\Application",
                    "Google\\Chrome Beta\\Application",
                    "Google\\Chrome Canary\\Application",
                )
                for base in [
                    os.getenv(env)
                    for env in (
                        "PROGRAMFILES",
                        "PROGRAMFILES(X86)",
                        "LOCALAPPDATA",
                    )
                ]
                if base is not None
            ]
            return ms_edge + google_chrome

        if is_darwin:
            return [
                "/Applications/Chromium.app/Contents/MacOS/Chromium",
                str(Path.home() / "Applications/Chromium.app/Contents/MacOS/Chromium"),
                "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome",
                str(Path.home() / "Applications/Google Chrome.app/Contents/MacOS/Google Chrome"),
            ]

        return []

    @classmethod
    def launch_args(cls) -> list[str]:
        # https://docs.google.com/spreadsheets/d/1n-vw_PCPS45jX3Jt9jQaAhFqBY6Ge1vWF_Pa0k7dCk4
        # https://peter.sh/experiments/chromium-command-line-switches/
        return [
            # Don't auto-play videos
            "--autoplay-policy=user-gesture-required",
            # Suppress all permission prompts by automatically denying them
            "--deny-permission-prompts",
            # Disable various background network services, including
            #   extension updating, safe browsing service, upgrade detector, translate, UMA
            "--disable-background-networking",
            # Chromium treats "foreground" tabs as "backgrounded" if the surrounding window is occluded by another window
            "--disable-backgrounding-occluded-windows",
            # Disable crashdump collection (reporting is already disabled in Chromium)
            "--disable-breakpad",
            # Disables client-side phishing detection
            "--disable-client-side-phishing-detection",
            # Disable some built-in extensions that aren't affected by `--disable-extensions`
            "--disable-component-extensions-with-background-pages",
            # Don't update the browser 'components' listed at chrome://components/
            "--disable-component-update",
            # Disable installation of default apps
            "--disable-default-apps",
            # Disable all chrome extensions
            "--disable-extensions",
            # Hide toolbar button that opens dialog for controlling media sessions
            "--disable-features=GlobalMediaControls",
            # Disable the "Chrome Media Router" which creates some background network activity to discover castable targets
            "--disable-features=MediaRouter",
            # Disables Chrome translation, both the manual option and the popup prompt
            "--disable-features=Translate",
            # Suppresses hang monitor dialogs in renderer processes
            #   This flag may allow slow unload handlers on a page to prevent the tab from closing
            "--disable-hang-monitor",
            # Disables logging
            "--disable-logging",
            # Disables the Web Notification and the Push APIs
            "--disable-notifications",
            # Disable popup blocking. `--block-new-web-contents` is the strict version of this
            "--disable-popup-blocking",
            # Reloading a page that came from a POST normally prompts the user
            "--disable-prompt-on-repost",
            # Disable syncing with Google
            "--disable-sync",
            # Forces the maximum disk space to be used by the disk cache, in bytes
            "--disk-cache-size=0",
            # Disable reporting to UMA, but allows for collection
            "--metrics-recording-only",
            # Mute any audio
            "--mute-audio",
            # Disable the default browser check, do not prompt to set it as such
            "--no-default-browser-check",
            # Disables all experiments set on about:flags
            "--no-experiments",
            # Skip first run wizards
            "--no-first-run",
            # Disables the service process from adding itself as an autorun process
            #   This does not delete existing autorun registrations, it just prevents the service from registering a new one
            "--no-service-autorun",
            # Avoid potential instability of using Gnome Keyring or KDE wallet
            "--password-store=basic",
            # No initial CDP target (no empty default tab)
            "--silent-launch",
            # Use mock keychain on Mac to prevent the blocking permissions dialog asking:
            #   Do you want the application "Chromium.app" to accept incoming network connections?
            "--use-mock-keychain",
            # When not using headless mode, try to disrupt the user as little as possible
            "--window-size=0,0",
        ]

    def __init__(
        self,
        *args,
        host: str | None = None,
        port: int | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.host = host or "127.0.0.1"
        self.port = port

    @asynccontextmanager
    async def launch(self, headless: bool = False, timeout: float | None = None) -> AsyncGenerator[trio.Nursery, None]:
        if self.port is None:
            if ":" in self.host:
                self.port = await find_free_port_ipv6(self.host)
            else:
                self.port = await find_free_port_ipv4(self.host)

        # no async rmtree
        with self._create_temp_dir() as user_data_dir:
            arguments = self.arguments.copy()
            if headless:
                arguments.append("--headless=new")
            arguments.extend([
                f"--remote-debugging-host={self.host}",
                f"--remote-debugging-port={self.port}",
                f"--user-data-dir={user_data_dir}",
            ])

            async with super()._launch(self.executable, arguments, headless=headless, timeout=timeout) as nursery:
                yield nursery

            # Even though we've awaited the process termination in the async generator above,
            # the rmtree() call of the temp-dir's context manager can sometimes still fail.
            # This is probably caused by filesystem commits of the OS, not sure,
            # but we have to wait a bit in order to be able to gracefully remove the temp user data dir.
            # A terrible solution to use a static timer :(
            await trio.sleep(0.5)

    def get_websocket_url(self, session: Streamlink) -> str:
        return session.http.get(
            f"http://{f'[{self.host}]' if ':' in self.host else self.host}:{self.port}/json/version",
            retries=10,
            retry_backoff=0.25,
            retry_max_backoff=0.25,
            timeout=0.1,
            proxies={
                "http": "",
            },
            schema=validate.Schema(
                validate.parse_json(),
                {"webSocketDebuggerUrl": validate.url(scheme="ws")},
                validate.get("webSocketDebuggerUrl"),
            ),
        )



from __future__ import annotations

import abc
import os
import random
import tempfile
import threading
from contextlib import suppress
from pathlib import Path

from streamlink.compat import is_win32
from streamlink.logger import getLogger


try:
    from ctypes import byref, c_ulong, c_void_p, cast, windll  # type: ignore[attr-defined]
except ImportError:
    pass


log = getLogger(__name__)

_lock = threading.Lock()
_id = 0


class NamedPipeBase(abc.ABC):
    path: Path

    def __init__(self):
        global _id  # noqa: PLW0603
        with _lock:
            _id += 1
            self.name = f"streamlinkpipe-{os.getpid()}-{_id}-{random.randint(0, 9999)}"
        log.info(f"Creating pipe {self.name}")
        self._create()

    @abc.abstractmethod
    def _create(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def open(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def write(self, data) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        raise NotImplementedError


class NamedPipePosix(NamedPipeBase):
    mode = "wb"
    permissions = 0o660
    fifo = None

    def _create(self):
        self.path = Path(tempfile.gettempdir(), self.name)
        os.mkfifo(self.path, self.permissions)

    def open(self):
        self.fifo = self.path.open(self.mode)

    def write(self, data):
        return self.fifo.write(data)  # type: ignore

    def close(self):
        try:
            if self.fifo is not None:
                self.fifo.close()
        except OSError:
            raise
        finally:
            with suppress(OSError):
                self.path.unlink()
            self.fifo = None


class NamedPipeWindows(NamedPipeBase):
    bufsize = 8192
    pipe = None

    PIPE_ACCESS_OUTBOUND = 0x00000002
    PIPE_TYPE_BYTE = 0x00000000
    PIPE_READMODE_BYTE = 0x00000000
    PIPE_WAIT = 0x00000000
    PIPE_UNLIMITED_INSTANCES = 255
    INVALID_HANDLE_VALUE = -1

    @staticmethod
    def _get_last_error():
        error_code = windll.kernel32.GetLastError()
        raise OSError(f"Named pipe error code 0x{error_code:08X}")

    def _create(self):
        self.path = Path("\\\\.\\pipe", self.name)
        self.pipe = windll.kernel32.CreateNamedPipeW(
            str(self.path),
            self.PIPE_ACCESS_OUTBOUND,
            self.PIPE_TYPE_BYTE | self.PIPE_READMODE_BYTE | self.PIPE_WAIT,
            self.PIPE_UNLIMITED_INSTANCES,
            self.bufsize,
            self.bufsize,
            0,
            None,
        )
        if self.pipe == self.INVALID_HANDLE_VALUE:
            self._get_last_error()

    def open(self):
        windll.kernel32.ConnectNamedPipe(self.pipe, None)

    def write(self, data):
        written = c_ulong(0)
        windll.kernel32.WriteFile(
            self.pipe,
            cast(data, c_void_p),
            len(data),
            byref(written),
            None,
        )
        return written.value

    def close(self):
        try:
            if self.pipe is not None:
                windll.kernel32.DisconnectNamedPipe(self.pipe)
                windll.kernel32.CloseHandle(self.pipe)
        except OSError:
            raise
        finally:
            self.pipe = None


NamedPipe: type[NamedPipeBase]
if not is_win32:
    NamedPipe = NamedPipePosix
else:
    NamedPipe = NamedPipeWindows



from __future__ import annotations

from textwrap import indent
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Sequence


class ValidationError(ValueError):
    """
    Currently not exposed in the public API.
    """

    MAX_LENGTH = 60

    errors: str | Exception | Sequence[str | Exception]

    def __init__(
        self,
        *errors,
        schema: str | object | None = None,
        **errkeywords,
    ):
        self.schema = schema
        if len(errors) == 1 and isinstance(errors[0], str):
            self.errors = (self._truncate(errors[0], **errkeywords),)
        else:
            self.errors = errors

    def _ellipsis(self, string: str):
        return string if len(string) <= self.MAX_LENGTH else f"<{string[: self.MAX_LENGTH - 5]}...>"

    def _truncate(self, template: str, **kwargs):
        return template.format(**{k: self._ellipsis(str(v)) for k, v in kwargs.items()})

    def _get_schema_name(self) -> str:
        if not self.schema:
            return ""
        if isinstance(self.schema, str):
            return f"({self.schema})"
        return f"({self.schema.__name__})"  # type: ignore[attr-defined, ty:unresolved-attribute]

    def __str__(self):
        cls = self.__class__
        ret = []
        seen = set()

        def append(indentation, error):
            if error:
                ret.append(indent(f"{error}", indentation))

        def add(level, error):
            indentation = "  " * level

            if error in seen:
                append(indentation, "...")
                return
            seen.add(error)

            for err in error.errors:
                if not isinstance(err, cls):
                    append(indentation, f"{err}")
                else:
                    append(indentation, f"{err.__class__.__name__}{err._get_schema_name()}:")
                    add(level + 1, err)

            context = error.__cause__
            if context:
                if not isinstance(context, cls):
                    append(indentation, "Context:")
                    append(f"{indentation}  ", context)
                else:
                    append(indentation, f"Context{context._get_schema_name()}:")
                    add(level + 1, context)

        append("", f"{cls.__name__}{self._get_schema_name()}:")
        add(1, self)

        return "\n".join(ret)



"""
$description Live TV channels and video on-demand service from RTP, a Portuguese public, state-owned broadcaster.
$url rtp.pt/play
$type live, vod
$region Portugal
"""

import re
from base64 import b64decode
from urllib.parse import unquote

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.stream.hls import HLSStream


@pluginmatcher(
    re.compile(r"https?://www\.rtp\.pt/play/"),
)
class RTPPlay(Plugin):
    def _get_streams(self):
        self.session.http.headers.update({
            "User-Agent": useragents.CHROME,
            "Referer": self.url,
        })

        re_m3u8 = re.compile(
            r"""
                hls\s*:\s*(?:
                    (?P<q>["'])(?P<string>.*?)(?P=q)
                    |
                    decodeURIComponent\s*\((?P<obfuscated>\[.*?])\.join\(
                    |
                    atob\s*\(\s*decodeURIComponent\s*\((?P<obfuscated_b64>\[.*?])\.join\(
                )
            """,
            re.VERBOSE,
        )

        hls_url = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.transform(lambda text: next(reversed(list(re_m3u8.finditer(text))), None)),
                validate.any(
                    None,
                    validate.all(
                        validate.get("string"),
                        str,
                        validate.any(
                            "",
                            validate.url(),
                        ),
                    ),
                    validate.all(
                        validate.get("obfuscated"),
                        str,
                        validate.parse_json(),
                        validate.transform(lambda arr: unquote("".join(arr))),
                        validate.url(),
                    ),
                    validate.all(
                        validate.get("obfuscated_b64"),
                        str,
                        validate.parse_json(),
                        validate.transform(lambda arr: unquote("".join(arr))),
                        validate.transform(lambda b64: b64decode(b64).decode("utf-8")),
                        validate.url(),
                    ),
                ),
            ),
        )
        if hls_url:
            return HLSStream.parse_variant_playlist(self.session, hls_url)


__plugin__ = RTPPlay



"""
$description Israeli live TV channel and video on-demand service owned by Network 13.
$url 13tv.co.il
$type live, vod
$region Israel
"""

import re
from urllib.parse import urljoin, urlunparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?13tv\.co\.il/(live|.*?/)"),
)
class N13TV(Plugin):
    api_url = "https://13tv-api.oplayer.io/api/getlink/"
    main_js_url_re = re.compile(r'type="text/javascript" src="(.*?main\..+\.js)"')
    user_id_re = re.compile(r'"data-ccid":"(.*?)"')
    video_name_re = re.compile(r'"videoRef":"(.*?)"')
    server_addr_re = re.compile(r"(.*[^/])(/.*)")
    media_file_re = re.compile(r"(.*)(\.[^\.].*)")

    live_schema = validate.Schema(
        [{"Link": validate.url()}],
        validate.get(0),
        validate.get("Link"),
    )

    vod_schema = validate.Schema(
        [
            {
                "ShowTitle": str,
                "ProtocolType": validate.all(
                    str,
                    validate.transform(lambda x: x.replace("://", "")),
                ),
                "ServerAddress": str,
                "MediaRoot": str,
                "MediaFile": str,
                "Bitrates": str,
                "StreamingType": str,
                "Token": validate.all(
                    str,
                    validate.transform(lambda x: x.lstrip("?")),
                ),
            },
        ],
        validate.get(0),
    )

    def _get_live(self, user_id):
        res = self.session.http.get(
            self.api_url,
            params=dict(
                userId=user_id,
                serverType="web",
                ch=1,
                cdnName="casttime",
            ),
        )

        url = self.session.http.json(res, schema=self.live_schema)
        log.debug(f"URL={url}")

        return HLSStream.parse_variant_playlist(self.session, url)

    def _get_vod(self, user_id, video_name):
        res = self.session.http.get(
            urljoin(self.api_url, "getVideoByFileName"),
            params=dict(
                userId=user_id,
                videoName=video_name,
                serverType="web",
                callback="x",
            ),
        )

        vod_data = self.session.http.json(res, schema=self.vod_schema)

        if video_name == vod_data["ShowTitle"]:
            try:
                host, base_path = self.server_addr_re.search(vod_data["ServerAddress"]).groups()  # type: ignore
            except AttributeError:
                raise PluginError("Could not split 'ServerAddress' components") from None

            try:
                base_file, file_ext = self.media_file_re.search(vod_data["MediaFile"]).groups()  # type: ignore
            except AttributeError:
                raise PluginError("Could not split 'MediaFile' components") from None

            media_path = "".join([
                base_path,
                vod_data["MediaRoot"],
                base_file,
                vod_data["Bitrates"],
                file_ext,
                vod_data["StreamingType"],
            ])
            log.debug("Media path=%s", media_path)

            vod_url = urlunparse((
                vod_data["ProtocolType"],
                host,
                media_path,
                "",
                vod_data["Token"],
                "",
            ))
            log.debug("URL=%s", vod_url)

            return HLSStream.parse_variant_playlist(self.session, vod_url)

    def _get_streams(self):
        url_type = self.match.group(1)
        log.debug(f"URL type={url_type}")

        res = self.session.http.get(self.url)

        if url_type != "live":
            m = self.video_name_re.search(res.text)
            video_name = m and m.group(1)
            if not video_name:
                raise PluginError("Could not determine video_name")
            log.debug(f"Video name={video_name}")

        m = self.main_js_url_re.search(res.text)
        main_js_path = m and m.group(1)
        if not main_js_path:
            raise PluginError("Could not determine main_js_path")
        log.debug(f"Main JS path={main_js_path}")

        res = self.session.http.get(urljoin(self.url, main_js_path))

        m = self.user_id_re.search(res.text)
        user_id = m and m.group(1)
        if not user_id:
            raise PluginError("Could not determine user_id")
        log.debug(f"User ID={user_id}")

        if url_type == "live":
            return self._get_live(user_id)
        else:
            return self._get_vod(user_id, video_name)


__plugin__ = N13TV



"""
$description Live TV channels and video on-demand service from MRT, a North Macedonian public, state-owned broadcaster.
$url play.mrt.com.mk
$type live, vod
$region North Macedonia
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://play\.mrt\.com\.mk/(live|play)/"),
)
class MRTmk(Plugin):
    file_re = re.compile(r"""(?P<url>https?://vod-[\d\w]+\.interspace\.com[^"',]+\.m3u8[^"',]*)""")

    stream_schema = validate.Schema(
        validate.all(
            validate.transform(file_re.finditer),
            validate.transform(list),
            [validate.get("url")],
            # remove duplicates
            validate.transform(set),
            validate.transform(list),
        ),
    )

    def _get_streams(self):
        res = self.session.http.get(self.url)
        stream_urls = self.stream_schema.validate(res.text)
        log.debug(f"Found streams: {len(stream_urls)}")
        if not stream_urls:
            return

        for stream_url in stream_urls:
            try:
                yield from HLSStream.parse_variant_playlist(self.session, stream_url).items()
            except OSError as err:
                if "403 Client Error" in str(err):
                    log.error("Failed to access stream, may be due to geo-restriction")
                else:
                    raise err


__plugin__ = MRTmk



from streamlink.session import Streamlink


def streams(url: str, **params):
    """
    Initializes an empty Streamlink session, attempts to find a plugin and extracts streams from the URL if a plugin was found.

    :param url: a URL to match against loaded plugins
    :param params: Additional keyword arguments passed to :meth:`Streamlink.streams() <streamlink.session.Streamlink.streams>`
    :raises NoPluginError: on plugin resolve failure
    :returns: A :class:`dict` of stream names and :class:`Stream <streamlink.stream.Stream>` instances
    """

    session = Streamlink()

    return session.streams(url, **params)



"""
$description Live TV channels and video on-demand service from SVT, a Swedish public, state-owned broadcaster.
$url svtplay.se
$type live, vod
$metadata author
$metadata category
$metadata title
$region Sweden
"""

import re
from urllib.parse import parse_qsl, urlparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.ffmpegmux import MuxedStream
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?svtplay\.se/(?P<live>kanaler/)?"),
)
class SVTPlay(Plugin):
    _URL_API_VIDEO = "https://api.svt.se/videoplayer-api/video/{item}"
    _MAP_CHANNEL_NAMES = {
        "svt1": "ch-svt1",
        "svt2": "ch-svt2",
        "svtbarn": "ch-barnkanalen",
        "kunskapskanalen": "ch-kunskapskanalen",
        "svt24": "ch-svt24",
    }

    def _api_call(self, item):
        schema_items = validate.all(
            [
                validate.all(
                    {
                        "format": str,
                        "url": validate.url(),
                    },
                    validate.union_get("format", "url"),
                ),
            ],
            validate.transform(dict),
        )

        self.author, self.title, videos, subtitles = self.session.http.get(
            self._URL_API_VIDEO.format(item=item),
            schema=validate.Schema(
                validate.parse_json(),
                {
                    validate.optional("programTitle"): str,
                    validate.optional("episodeTitle"): str,
                    "videoReferences": schema_items,
                    validate.optional("subtitleReferences"): schema_items,
                },
                validate.union_get(
                    "programTitle",
                    "episodeTitle",
                    "videoReferences",
                    "subtitleReferences",
                ),
            ),
        )

        return videos, subtitles

    def _get_live(self):
        live_id = "/".join(urlparse(self.url).path.split("/")[2:])
        if not live_id:
            return

        live_id = self._MAP_CHANNEL_NAMES.get(live_id, f"ch-{live_id}")
        log.debug(f"Live ID={live_id}")
        self.category = "Live"
        videos, subtitles = self._api_call(live_id)

        return self._select_streams(videos, subtitles)

    def _get_vod(self):
        def get_vod_id(url):
            return dict(parse_qsl(urlparse(url).query)).get("id")

        vod_id = get_vod_id(self.url)

        if vod_id is None:
            vod_url = self.session.http.get(
                self.url,
                schema=validate.Schema(
                    validate.parse_html(),
                    validate.xml_xpath_string(".//*[@data-rt='top-area-play-button'][@href][1]/@href"),
                ),
            )
            if vod_url:
                vod_id = get_vod_id(vod_url)

        if vod_id is None:
            return

        log.debug(f"VOD ID={vod_id}")
        self.category = "Live"
        videos, subtitles = self._api_call(vod_id)

        return self._select_streams(videos, subtitles)

    def _select_streams(self, videos, subtitles):
        # the goal is to have streams with the widest range of qualities/substreams and highest bitrate at the top
        stream_priorities = {
            "dashhbbtv": DASHStream,  # DASH AVC
            "dash-hbbtv-avc": DASHStream,  # DASH AVC
            "dash-avc": DASHStream,  # DASH AVC
            "dash-full": DASHStream,  # DASH AVC
            "dash": DASHStream,  # DASH AVC
            "hlswebvtt": HLSStream,  # HLS with subtitles
            "hls-cmaf-live-vtt": HLSStream,  # HLS with subtitles
            "hls-ts-avc": HLSStream,  # HLS with MPEG-TS
            "hls-ts-full": HLSStream,  # HLS with MPEG-TS
            "hls": HLSStream,  # HLS with MPEG-TS
            "hls-cmaf-live": HLSStream,  # HLS with fMP4
            "hls-cmaf-full": HLSStream,  # HLS with fMP4
            "dash-hbbtv-hevc": DASHStream,  # DASH HEVC (low prio, because of potential user decoder issues)
            "hls-ts-lb-full": HLSStream,  # low bitrate
            "hls-cmaf-lb-full": HLSStream,  # low bitrate
            "dash-lb-full": DASHStream,  # low bitrate
        }

        for fmt, streamtype in stream_priorities.items():
            if fmt not in videos:
                continue

            if streamtype is HLSStream:
                return HLSStream.parse_variant_playlist(self.session, videos[fmt], name_fmt="{pixels}_{bitrate}")

            if streamtype is DASHStream:
                subtitlestreams = {}
                if self.session.get_option("mux-subtitles") and "webvtt" in subtitles:
                    subtitlestreams["webvtt"] = HTTPStream(self.session, subtitles["webvtt"])

                dash_streams = DASHStream.parse_manifest(self.session, videos[fmt])
                if not subtitlestreams:
                    return dash_streams

                return {q: MuxedStream(self.session, s, subtitles=subtitlestreams) for q, s in dash_streams.items()}

    def _get_streams(self):
        if self.match["live"]:
            return self._get_live()
        else:
            return self._get_vod()


__plugin__ = SVTPlay



from __future__ import annotations

import re
from typing import Any, Generic, TypeVar


_BOOLEAN_TRUE = "yes", "1", "true", "on"
_BOOLEAN_FALSE = "no", "0", "false", "off"

_FILESIZE_RE = re.compile(r"^(?P<size>\d+(\.\d+)?)(?P<modifier>[km])?b?$", re.IGNORECASE)
_FILESIZE_UNITS = {
    "k": 2**10,
    "m": 2**20,
}

_KEYVALUE_RE = re.compile(r"^(?P<key>[^=\s]+)\s*=\s*(?P<value>.*)$")


def boolean(value: str) -> bool:
    if value.lower() not in _BOOLEAN_TRUE + _BOOLEAN_FALSE:
        raise ValueError(f"{value} is not one of {{{', '.join(_BOOLEAN_TRUE + _BOOLEAN_FALSE)}}}")

    return value.lower() in _BOOLEAN_TRUE


def comma_list(values: str) -> list[str]:
    return [val.strip() for val in values.split(",")]


# noinspection PyPep8Naming
class comma_list_filter:
    def __init__(self, acceptable: list[str], unique: bool = False):
        self.acceptable = tuple(acceptable)
        self.unique = unique

    def __call__(self, values: str) -> list[str]:
        res = [item for item in comma_list(values) if item in self.acceptable]
        return sorted(set(res)) if self.unique else res

    def __hash__(self):
        return hash((self.acceptable, self.unique))


def filesize(value: str) -> int:
    match = _FILESIZE_RE.match(value.strip())
    if not match:
        raise ValueError("Invalid file size format")

    size = float(match["size"])
    size *= _FILESIZE_UNITS.get((match["modifier"] or "").lower(), 1)

    return num(int, ge=1)(size)


def keyvalue(value: str) -> tuple[str, str]:
    match = _KEYVALUE_RE.match(value.lstrip())
    if not match:
        raise ValueError("Invalid key=value format")

    return match["key"], match["value"]


_TNum = TypeVar("_TNum", int, float)


# noinspection PyPep8Naming
class num(Generic[_TNum]):
    def __init__(
        self,
        numtype: type[_TNum],
        ge: _TNum | None = None,
        gt: _TNum | None = None,
        le: _TNum | None = None,
        lt: _TNum | None = None,
    ):
        self.numtype: type[_TNum] = numtype
        self.ge: _TNum | None = ge
        self.gt: _TNum | None = gt
        self.le: _TNum | None = le
        self.lt: _TNum | None = lt
        self.__name__ = numtype.__name__

    def __call__(self, value: Any) -> _TNum:
        val: _TNum = self.numtype(value)

        if self.ge is not None and not (val >= self.ge):
            raise ValueError(f"{self.__name__} value must be >={self.ge}, but is {val}")
        if self.gt is not None and not (val > self.gt):
            raise ValueError(f"{self.__name__} value must be >{self.gt}, but is {val}")
        if self.le is not None and not (val <= self.le):
            raise ValueError(f"{self.__name__} value must be <={self.le}, but is {val}")
        if self.lt is not None and not (val < self.lt):
            raise ValueError(f"{self.__name__} value must be <{self.lt}, but is {val}")

        return val

    def __hash__(self) -> int:
        return hash((self.numtype, self.ge, self.gt, self.le, self.lt))


__all__ = [
    "boolean",
    "comma_list",
    "comma_list_filter",
    "filesize",
    "keyvalue",
    "num",
]



"""
$description Spanish live TV channel for Telemadrid, a public regional television station.
$url telemadrid.es
$type live, vod
$region Spain
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.plugins.brightcove import BrightcovePlayer


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?telemadrid\.es/"),
)
class Telemadrid(Plugin):
    def _get_streams(self):
        data = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_find(".//video-js[@data-video-id][@data-account][@data-player][1]"),
                validate.union_get("data-video-id", "data-account", "data-player"),
            ),
        )
        data_video_id, data_account, data_player = data
        player = BrightcovePlayer(self.session, data_account, f"{data_player}_default")
        return player.get_streams(data_video_id)


__plugin__ = Telemadrid



"""
$description Stream videos stored in Google Drive.
$url docs.google.com
$url drive.google.com
$type vod
"""

import re
from urllib.parse import parse_qsl

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.stream.http import HTTPStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:drive|docs)\.google\.com/file/d/([^/]+)/?"),
)
class GoogleDocs(Plugin):
    api_url = "https://docs.google.com/get_video_info"

    def _get_streams(self):
        docid = self.match.group(1)
        log.debug("Google Docs ID: %s", docid)
        res = self.session.http.get(self.api_url, params=dict(docid=docid))
        data = dict(parse_qsl(res.text))

        if data["status"] == "ok":
            fmts = dict([s.split("/")[:2] for s in data["fmt_list"].split(",")])
            streams = [s.split("|") for s in data["fmt_stream_map"].split(",")]
            for qcode, url in streams:
                _, h = fmts[qcode].split("x")
                yield f"{h}p", HTTPStream(self.session, url)
        else:
            log.error("%s (ID: %s)", data["reason"], docid)


__plugin__ = GoogleDocs



"""
$description British live TV channel and video on-demand service from Blaze, owned by A&E Networks UK.
$url blaze.tv
$type live, vod
$metadata id
$metadata author
$metadata category
$metadata title
$region United Kingdom
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:watch\.)?blaze\.tv/(?:(?P<is_live>live)|watch/replay/\d+)"),
)
class BlazeTV(Plugin):
    @staticmethod
    def _get_live_uvid(parsed_html):
        schema = validate.Schema(
            validate.xml_xpath_string(".//div[@id='live-player-root']/@data-player-uvid"),
        )
        return schema.validate(parsed_html)

    @staticmethod
    def _get_vod_uvid(parsed_html):
        schema = validate.Schema(
            validate.xml_xpath_string(".//script[contains(text(), 'window.nowPlaying.setData')]"),
            validate.none_or_all(
                re.compile(r"window\.nowPlaying\.setData\(({.*?})\);"),
                validate.none_or_all(
                    validate.get(1),
                    validate.parse_json(),
                    {
                        "id": str,
                        "series_title": str,
                        "title": str,
                        "season": str,
                        "episode": str,
                    },
                ),
            ),
        )
        return schema.validate(parsed_html)

    def _get_tokenizer(self, streamtype, uvid):
        return self.session.http.get(
            f"https://watch.blaze.tv/stream/{streamtype}/widevine/{uvid}",
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "tokenizer": {
                        "url": validate.url(),
                        "uvid": str,
                        "expiry": int,
                        "token": str,
                    },
                },
                validate.get("tokenizer"),
            ),
        )

    def _get_streams(self):
        is_live = self.match.group("is_live")
        parsed_html = self.session.http.get(self.url, schema=validate.Schema(validate.parse_html()))

        if is_live:
            uvid = self._get_live_uvid(parsed_html)
            if not uvid or not uvid.isdecimal():
                return
            token_data = self._get_tokenizer("live", uvid)
            self.id = uvid
            self.author = "Blaze"
            self.title = "Live TV"
            self.category = "Live"
        else:
            data = self._get_vod_uvid(parsed_html)
            if not data or not data["id"] or not data["id"].isdecimal():
                return
            token_data = self._get_tokenizer("replay", data["id"])
            self.id = data["id"]
            self.author = data["series_title"]
            self.title = data["title"]
            self.category = f"S{data['season']}E{data['episode']}"

        log.trace("token_data=%r", token_data)

        hls_url = self.session.http.get(
            token_data["url"],
            headers={
                "Token": token_data["token"],
                "Token-Expiry": str(token_data["expiry"]),
                "Uvid": token_data["uvid"],
            },
            schema=validate.Schema(
                validate.parse_json(),
                {"Streams": {"Adaptive": validate.url()}},
                validate.get(("Streams", "Adaptive")),
            ),
        )
        return HLSStream.parse_variant_playlist(self.session, hls_url)


__plugin__ = BlazeTV



"""
$description French free-to-air news channel, providing 24-hour national and global news coverage.
$url cnews.fr
$type live, vod
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?cnews\.fr"),
)
class CNEWS(Plugin):
    _dailymotion_url = "https://www.dailymotion.com/embed/video/{}"

    def _get_streams(self):
        data = self.session.http.get(
            self.url,
            schema=validate.Schema(
                re.compile(r"jQuery\.extend\(Drupal\.settings, ({.*})\);"),
                validate.none_or_all(
                    validate.get(1),
                    validate.parse_json(),
                    {
                        validate.optional("dm_player_live_dailymotion"): {
                            validate.optional("video_id"): str,
                        },
                        validate.optional("dm_player_node_dailymotion"): {
                            validate.optional("video_id"): str,
                        },
                    },
                    validate.union_get("dm_player_live_dailymotion", "dm_player_node_dailymotion"),
                ),
            ),
        )
        if not data:
            return

        live, node = data

        if node:
            return self.session.streams(self._dailymotion_url.format(node["video_id"]))
        elif live:
            return self.session.streams(self._dailymotion_url.format(live["video_id"]))


__plugin__ = CNEWS



"""
$description Live TV channels from CT, a Czech public, state-owned broadcaster.
$url ceskatelevize.cz
$type live
$metadata id
$metadata title
$region Czechia
"""

import re
from uuid import uuid4

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream


log = getLogger(__name__)


@pluginmatcher(
    name="channel",
    pattern=re.compile(r"https?://(?P<channel>ct24|decko)\.ceskatelevize\.cz/"),
)
@pluginmatcher(
    name="sport",
    pattern=re.compile(r"https?://sport\.ceskatelevize\.cz/"),
)
@pluginmatcher(
    name="default",
    pattern=re.compile(r"https?://(?:www\.)?ceskatelevize\.cz/zive/\w+"),
)
class Ceskatelevize(Plugin):
    URL_API_CHANNEL = "https://api.ceskatelevize.cz/video/v1/playlist-live/v1/stream-data/channel/{channel}"

    CHANNELS = {
        "ct24": "CH_24",
        "decko": "CH_5",
    }

    def get_streams_from_api(self, channel):
        if not channel:
            return

        self.title, is_blocked, url = self.session.http.get(
            self.URL_API_CHANNEL.format(channel=channel),
            params={
                "canPlayDRM": "false",
                "streamType": "dash",
                "quality": "web",
                "maxQualityCount": 5,
                "sessionId": str(uuid4()),
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "streamUrls": {
                        "isBlocked": bool,
                        "main": validate.url(),
                    },
                    "title": str,
                },
                validate.union_get(
                    "title",
                    ("streamUrls", "isBlocked"),
                    ("streamUrls", "main"),
                ),
            ),
        )

        if is_blocked:
            log.error("The stream is inaccessible")
            return

        return url

    def get_streams_channel(self):
        self.id = self.CHANNELS.get(self.match["channel"])

        return self.get_streams_from_api(self.id)

    def get_streams_sport(self):
        schema = validate.Schema(
            validate.parse_html(),
            validate.xml_xpath_string(".//script[@id='__NEXT_DATA__'][text()][1]/text()"),
            validate.none_or_all(
                str,
                validate.parse_json(),
                {
                    "props": {
                        "pageProps": {
                            "liveData": validate.any(
                                # Prefer explicit broadcast id (eg "CH_4")
                                validate.all({"id": str}, validate.get("id")),
                                # Fallback to encoder (also "CH_4")
                                validate.all(
                                    {"current": {"encoder": str}},
                                    validate.get(("current", "encoder")),
                                ),
                            ),
                        },
                    },
                },
                validate.get(("props", "pageProps", "liveData")),
            ),
        )

        self.id = self.session.http.get(self.url, schema=schema)

        return self.get_streams_from_api(self.id)

    def get_streams_default(self):
        schema = validate.Schema(
            validate.parse_html(),
            validate.xml_xpath_string(".//script[@id='__NEXT_DATA__'][text()]/text()"),
            str,
            validate.parse_json(),
            {
                "props": {
                    "pageProps": {
                        validate.optional("data"): validate.all(
                            {
                                "liveBroadcast": {
                                    "id": str,
                                },
                            },
                            validate.get(("liveBroadcast", "id")),
                        ),
                    },
                },
            },
            validate.get(("props", "pageProps", "data")),
        )

        self.id = self.session.http.get(self.url, schema=schema)

        return self.get_streams_from_api(self.id)

    def _get_streams(self):
        if self.matches["channel"]:
            url = self.get_streams_channel()
        elif self.matches["sport"]:
            url = self.get_streams_sport()
        else:
            url = self.get_streams_default()

        if not url:
            return

        return DASHStream.parse_manifest(self.session, url)


__plugin__ = Ceskatelevize



"""
$description French live TV channels, live streams and video content, owned by NextRadioTV.
$url bfmtv.com
$url 01net.com
$type live, vod
"""

import re
from urllib.parse import urljoin

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.plugins.brightcove import BrightcovePlayer
from streamlink.stream.http import HTTPStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:[\w-]+\.)+(?:bfmtv|01net)\.com"),
)
class BFMTV(Plugin):
    def _brightcove(self, account_id, video_id):
        log.debug(f"Account ID: {account_id}")
        log.debug(f"Video ID: {video_id}")
        player = BrightcovePlayer(self.session, account_id)

        return dict(player.get_streams(video_id))

    def _streams_brightcove(self, root):
        schema_brightcove = validate.Schema(
            validate.any(
                validate.all(
                    validate.xml_find(".//*[@accountid][@videoid]"),
                    validate.union_get("accountid", "videoid"),
                ),
                validate.all(
                    validate.xml_find(".//*[@data-account][@data-video-id]"),
                    validate.union_get("data-account", "data-video-id"),
                ),
            ),
        )
        try:
            account_id, video_id = schema_brightcove.validate(root)
        except PluginError:
            return

        return self._brightcove(account_id, video_id)

    def _streams_brightcove_js(self, root):
        re_js_src = re.compile(r"^[\w/]+/main\.\w+\.js$")
        schema_brightcove_js = validate.Schema(
            validate.xml_findall(r".//script[@src]"),
            validate.filter(lambda elem: re_js_src.search(elem.attrib.get("src")) is not None),
            validate.get(0),
            str,
            validate.transform(lambda src: urljoin(self.url, src)),
        )
        schema_brightcove_js2 = validate.Schema(
            re.compile(r"""i\?\([A-Z]="[^"]+",y="(?P<video_id>\d+).*"data-account"\s*:\s*"(?P<account_id>\d+)"""),
            validate.union_get("account_id", "video_id"),
        )
        try:
            js_url = schema_brightcove_js.validate(root)
            log.debug(f"JS URL: {js_url}")
            account_id, video_id = self.session.http.get(js_url, schema=schema_brightcove_js2)
        except PluginError:
            return

        return self._brightcove(account_id, video_id)

    def _streams_dailymotion(self, root):
        schema_dailymotion = validate.Schema(
            validate.xml_xpath_string(".//iframe[contains(@src,'dailymotion.com/')][1]/@src"),
            str,
            validate.transform(lambda src: src.split("/")[-1]),
        )
        try:
            video_id = schema_dailymotion.validate(root)
        except PluginError:
            return

        log.debug(f"Found dailymotion video ID: {video_id}")

        return self.session.streams(f"https://www.dailymotion.com/embed/video/{video_id}")

    def _streams_audio(self, root):
        schema_audio = validate.Schema(
            validate.any(
                validate.all(
                    validate.xml_xpath_string(".//audio/source[contains(@src,'.mp3')][1]/@src"),
                    str,
                ),
                validate.all(
                    validate.xml_xpath_string(".//div[contains(@class,'audio-player')][@data-media-url][1]/@data-media-url"),
                    str,
                ),
            ),
        )
        try:
            audio_url = schema_audio.validate(root)
        except PluginError:
            return

        return {"audio": HTTPStream(self.session, audio_url)}

    def _get_streams(self):
        root = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
            ),
        )

        return (
            self._streams_brightcove(root)
            or self._streams_dailymotion(root)
            or self._streams_brightcove_js(root)
            or self._streams_audio(root)
        )


__plugin__ = BFMTV



from __future__ import annotations

from threading import Event
from typing import TYPE_CHECKING

from streamlink.stream.stream import StreamIO


if TYPE_CHECKING:
    from streamlink.buffers import Buffer


class FilteredStream(StreamIO):
    """StreamIO mixin for being able to pause read calls while filtering content"""

    buffer: Buffer

    def __init__(self, *args, **kwargs):
        self._event_filter = Event()
        self._event_filter.set()
        super().__init__(*args, **kwargs)

    def read(self, *args, **kwargs) -> bytes:
        read = super().read
        while True:
            try:
                return read(*args, **kwargs)
            except OSError:
                # wait indefinitely until filtering ends
                self._event_filter.wait()
                if self.buffer.closed:
                    return b""
                # if data is available, try reading again
                if self.buffer.length > 0:
                    continue
                # raise if not filtering and no data available
                raise

    def close(self) -> None:
        super().close()
        self._event_filter.set()

    def is_paused(self) -> bool:
        return not self._event_filter.is_set()

    def pause(self) -> None:
        self._event_filter.clear()

    def resume(self) -> None:
        self._event_filter.set()

    def filter_wait(self, timeout=None):
        return self._event_filter.wait(timeout)



"""
$description Japanese live-streaming and video hosting social platform.
$url live.nicovideo.jp
$type live, vod
$metadata id
$metadata author
$metadata title
$account Required by some streams
$notes Timeshift is supported
"""

from __future__ import annotations

import re
from threading import Event
from typing import TYPE_CHECKING
from urllib.parse import urljoin

from requests.exceptions import HTTPError

from streamlink.exceptions import StreamError
from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.plugin.api.websocket import WebsocketClient
from streamlink.stream.hls import HLSStream, HLSStreamReader, HLSStreamWriter
from streamlink.utils.parse import parse_json
from streamlink.utils.url import update_qsd


if TYPE_CHECKING:
    from streamlink.stream.hls import HLSSegment


log = getLogger(__name__)


class NicoLiveWsClient(WebsocketClient):
    STREAM_OPENED_TIMEOUT = 6

    ready: Event
    opened: Event
    hls_stream_url: str

    _SCHEMA_COOKIES = validate.Schema(
        [
            {
                "domain": str,
                "path": str,
                "name": str,
                "value": str,
                "secure": bool,
            },
        ],
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.opened = Event()
        self.ready = Event()

    def reconnect(self, *args, **kwargs):
        self.ready.clear()
        return super().reconnect(*args, **kwargs)

    def on_open(self, wsapp):
        super().on_open(wsapp)
        self.send_playerversion()
        self.send_getpermit()

    def on_message(self, wsapp, data: str):
        log.debug(f"Received: {data}")
        message = parse_json(data)
        msgtype = message.get("type")
        msgdata = message.get("data", {})

        if handler := self._MESSAGE_HANDLERS.get(msgtype):
            handler(self, msgdata)

    def on_message_ping(self, _data):
        self.send_pong()

    def on_message_disconnect(self, data):
        reason = data.get("reason", "Unknown reason")
        log.info(f"Received disconnect message: {reason}")
        self.close()

    def on_message_stream(self, data):
        if data.get("protocol") != "hls" or not data.get("uri"):
            return

        # cookies may be required by some HLS multivariant playlists
        if cookies := data.get("cookies", []):
            log.info("Applying HTTP session cookies from websocket data")
            for cookie in self._SCHEMA_COOKIES.validate(cookies):
                self.session.http.cookies.set(**cookie)

        self.hls_stream_url = data.get("uri")
        self.ready.set()
        if self.opened.wait(self.STREAM_OPENED_TIMEOUT):
            log.debug("Stream opened, keeping websocket connection alive")
        else:
            log.info("Closing websocket connection")
            self.close()

    _MESSAGE_HANDLERS = {
        "ping": on_message_ping,
        "disconnect": on_message_disconnect,
        "stream": on_message_stream,
    }

    def send_playerversion(self):
        self.send_json({
            "type": "startWatching",
            "data": {
                "stream": {
                    "quality": "abr",
                    "protocol": "hls",
                    "latency": "high",
                    "chasePlay": False,
                },
                "room": {
                    "protocol": "webSocket",
                    "commentable": True,
                },
                "reconnect": False,
            },
        })

    def send_getpermit(self):
        self.send_json({
            "type": "getAkashic",
            "data": {
                "chasePlay": False,
            },
        })

    def send_pong(self):
        self.send_json({"type": "pong"})
        self.send_json({"type": "keepSeat"})


class NicoLiveHLSStreamWriter(HLSStreamWriter):
    reader: NicoLiveHLSStreamReader
    stream: NicoLiveHLSStream

    def create_decryptor(self, *args, **kwargs):
        try:
            return super().create_decryptor(*args, **kwargs)
        except StreamError as err:
            # TODO: fix HTTPSession.request()
            if (
                not (orig_err := getattr(err, "err", None))
                or not isinstance(orig_err, HTTPError)
                or orig_err.response.status_code != 403
            ):
                raise

            if not self.stream.wsclient.is_reconnecting.is_set():
                log.warning("HLSSegment decryption key retrieval failed. Attempting to reconnect to websocket...")
                self.stream.wsclient.reconnect()
            if (
                not self.stream.wsclient.reconnect_done.wait(self.stream.wsclient.STREAM_OPENED_TIMEOUT)
                or not self.stream.wsclient.ready.wait(self.stream.wsclient.STREAM_OPENED_TIMEOUT)
            ):  # fmt: skip
                raise

            # drop cached key
            self.key_uri = None

            return super().create_decryptor(*args, **kwargs)

    def should_filter_segment(self, segment: HLSSegment) -> bool:
        if "/blank/" in segment.uri:
            return True

        return super().should_filter_segment(segment)


class NicoLiveHLSStreamReader(HLSStreamReader):
    __writer__ = NicoLiveHLSStreamWriter

    writer: NicoLiveHLSStreamWriter
    stream: NicoLiveHLSStream

    def open(self):
        self.stream.wsclient.opened.set()
        super().open()

    def close(self):
        super().close()
        self.stream.wsclient.close()


class NicoLiveHLSStream(HLSStream):
    __reader__ = NicoLiveHLSStreamReader
    wsclient: NicoLiveWsClient

    def __init__(self, *args, wsclient: NicoLiveWsClient, **kwargs):
        super().__init__(*args, **kwargs)
        self.wsclient = wsclient


@pluginmatcher(
    re.compile(r"https?://(?P<domain>live\d*\.nicovideo\.jp)/watch/(lv|co|user/)\d+"),
)
@pluginargument(
    "email",
    sensitive=True,
    argument_name="niconico-email",
    metavar="EMAIL",
    help="The email or phone number associated with your Niconico account",
)
@pluginargument(
    "password",
    sensitive=True,
    argument_name="niconico-password",
    metavar="PASSWORD",
    help="The password of your Niconico account",
)
@pluginargument(
    "user-session",
    sensitive=True,
    argument_name="niconico-user-session",
    metavar="VALUE",
    help="""
        Value of the user-session token.

        Can be used as an alternative to providing a password.
    """,
)
@pluginargument(
    "purge-credentials",
    argument_name="niconico-purge-credentials",
    action="store_true",
    help="Purge cached Niconico credentials to initiate a new session and reauthenticate.",
)
@pluginargument(
    "timeshift-offset",
    type="hours_minutes_seconds",
    argument_name="niconico-timeshift-offset",
    metavar="[[XX:]XX:]XX | [XXh][XXm][XXs]",
    help="""
        Amount of time to skip from the beginning of a stream.

        Default is 0.
    """,
)
class NicoLive(Plugin):
    STREAM_READY_TIMEOUT = 6
    LOGIN_URL = "https://account.nicovideo.jp/login/redirector"
    LOGIN_URL_PARAMS = {
        "site": "niconico",
    }

    wsclient: NicoLiveWsClient

    def _get_streams(self):
        if self.get_option("purge_credentials"):
            self.clear_cookies()
            log.info("All credentials were successfully removed")

        self.session.http.headers.update({
            "User-Agent": useragents.CHROME,
        })

        self.niconico_web_login()

        data = self.get_data()

        wss_api_url = self.find_wss_api_url(data)
        if not wss_api_url:
            log.error(
                "Failed to get wss_api_url. "
                + "Please check if the URL is correct, and make sure your account has access to the video.",
            )
            return

        self.id, self.author, self.title = self.find_metadata(data)

        self.wsclient = NicoLiveWsClient(self.session, wss_api_url)
        self.wsclient.start()

        hls_stream_url = self._get_hls_stream_url()
        if not hls_stream_url:
            return

        offset = self.get_option("timeshift-offset")
        if offset and "timeshift" in wss_api_url:
            hls_stream_url = update_qsd(hls_stream_url, {"start": offset})

        return NicoLiveHLSStream.parse_variant_playlist(
            self.session,
            hls_stream_url,
            wsclient=self.wsclient,
            ffmpeg_options={"copyts": True},
        )

    def _get_hls_stream_url(self):
        log.debug(f"Waiting for permit (for at most {self.STREAM_READY_TIMEOUT} seconds)...")
        if not self.wsclient.ready.wait(self.STREAM_READY_TIMEOUT) or not self.wsclient.is_alive():
            log.error("Waiting for permit timed out.")
            self.wsclient.close()
            return

        return self.wsclient.hls_stream_url

    def get_data(self):
        return self.session.http.get(
            self.url,
            encoding="utf-8",
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_find(".//script[@id='embedded-data'][@data-props]"),
                validate.get("data-props"),
                validate.parse_json(),
            ),
        )

    @staticmethod
    def find_metadata(data):
        schema = validate.Schema(
            {
                "program": {
                    "nicoliveProgramId": str,
                    "supplier": {"name": str},
                    "title": str,
                },
            },
            validate.get("program"),
            validate.union_get(
                "nicoliveProgramId",
                ("supplier", "name"),
                "title",
            ),
        )

        return schema.validate(data)

    @staticmethod
    def find_wss_api_url(data):
        schema = validate.Schema(
            {
                "site": {
                    "relive": {
                        "webSocketUrl": validate.any(
                            validate.url(scheme="wss"),
                            "",
                        ),
                    },
                    validate.optional("frontendId"): int,
                },
            },
            validate.get("site"),
            validate.union_get(("relive", "webSocketUrl"), "frontendId"),
        )

        wss_api_url, frontend_id = schema.validate(data)
        if not wss_api_url:
            return

        if frontend_id is not None:
            wss_api_url = update_qsd(wss_api_url, {"frontend_id": frontend_id})

        return wss_api_url

    def niconico_web_login(self):
        user_session = self.get_option("user-session")
        email = self.get_option("email")
        password = self.get_option("password")

        if user_session is not None:
            log.info("Logging in via provided user session cookie")
            self.session.http.cookies.set(
                "user_session",
                user_session,
                path="/",
                domain="nicovideo.jp",
            )
            self.save_cookies()

        elif self.session.http.cookies.get("user_session"):
            log.info("Logging in via cached user session cookie")

        elif email is not None and password is not None:
            log.info("Logging in via provided email and password")
            root = self.session.http.post(
                self.LOGIN_URL,
                data={"mail_tel": email, "password": password},
                params=self.LOGIN_URL_PARAMS,
                schema=validate.Schema(validate.parse_html()),
            )

            if self.session.http.cookies.get("user_session"):
                log.info("Logged in.")
                self.save_cookies()
                return

            input_with_value = {}
            for elem in root.xpath(".//form[@action]//input"):
                if elem.attrib.get("value"):
                    input_with_value[elem.attrib.get("name")] = elem.attrib.get("value")
                elif elem.attrib.get("id") == "oneTimePw":
                    maxlength = int(elem.attrib.get("maxlength"))
                    oneTimePw = self.input_ask("Enter the 6 digit number included in email")
                    if len(oneTimePw) > maxlength:
                        log.error("invalid user input")
                        return
                    input_with_value[elem.attrib.get("name")] = oneTimePw
                else:
                    log.debug(f"unknown input: {elem.attrib.get('name')}")

            root = self.session.http.post(
                urljoin("https://account.nicovideo.jp", root.xpath("string(.//form[@action]/@action)")),
                data=input_with_value,
                schema=validate.Schema(validate.parse_html()),
            )
            log.debug(f"Cookies: {self.session.http.cookies.get_dict()}")
            if self.session.http.cookies.get("user_session") is None:
                error = root.xpath("string(//div[@class='formError']/div/text())")
                log.warning(f"Login failed: {error or 'unknown reason'}")
            else:
                log.info("Logged in.")
                self.save_cookies()


__plugin__ = NicoLive



"""
$description Global live-streaming and video on-demand hosting platform.
$url ott.streann.com
$url centroecuador.ec
$url columnaestilos.com
$url evtv.online/noticias-de-venezuela
$url telecuracao.com
$type live, vod
$metadata title
"""

import base64
import random
import re
import time
from urllib.parse import urlparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.crypto import decrypt_openssl
from streamlink.utils.parse import parse_qsd


log = getLogger(__name__)


@pluginmatcher(
    name="centroecuador",
    pattern=re.compile(r"https?://(?:www\.)?centroecuador\.ec"),
)
@pluginmatcher(
    name="columnaestilos",
    pattern=re.compile(r"https?://(?:www\.)?columnaestilos\.com"),
)
@pluginmatcher(
    name="crc",
    pattern=re.compile(r"https?://(?:www\.)?crc\.cr/estaciones/"),
)
@pluginmatcher(
    name="evtv",
    pattern=re.compile(r"https?://(?:www\.)?evtv\.online/noticias-de-venezuela/"),
)
@pluginmatcher(
    name="telecuracao",
    pattern=re.compile(r"https?://(?:www\.)?telecuracao\.com"),
)
@pluginmatcher(
    name="streann",
    pattern=re.compile(r"https?://ott\.streann\.com/s(?:treaming|-secure)/player\.html"),
)
@pluginargument(
    "url",
    metavar="URL",
    help="Source URL where the iframe is located, only required for direct URLs of ott.streann.com",
)
class Streann(Plugin):
    base_url = "https://ott.streann.com"
    get_time_url = base_url + "/web/services/public/get-server-time"
    token_url = base_url + "/loadbalancer/services/web-players/{playerId}/token/{type}/{dataId}/{deviceId}"
    stream_url = (
        base_url
        + "/loadbalancer/services/web-players/{type}s-reseller-secure/{dataId}/{playerId}"
        + "/{token}/{resellerId}/playlist.m3u8?date={time}&device-type=web&device-name=web"
        + "&device-os=web&device-id={deviceId}"
    )
    passphrase_re = re.compile(r"""CryptoJS\.AES\.decrypt\(.*?,\s*(['"])(?P<passphrase>(?:(?!\1).)*)\1\s*?\);""")

    _device_id = None
    _domain = None

    @property
    def device_id(self):
        """
        Randomly generated deviceId.
        :return:
        """
        if self._device_id is None:
            self._device_id = "".join(
                random.choice("0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ") for _ in range(50)
            )
        return self._device_id

    @property
    def time(self):
        res = self.session.http.get(self.get_time_url)
        data = self.session.http.json(res)
        return str(data.get("serverTime", int(time.time() * 1000)))

    def passphrase(self):
        log.debug("passphrase")
        res = self.session.http.get(self.url)
        passphrase_m = self.passphrase_re.search(res.text)
        return passphrase_m and passphrase_m.group("passphrase").encode("utf8")

    def get_token(self, **config):
        log.debug("get_token")
        pdata = dict(
            arg1=base64.b64encode(self._domain.encode("utf8")),  # type: ignore
            arg2=base64.b64encode(self.time.encode("utf8")),
        )

        headers = {
            "Referer": self.url,
            "X-Requested-With": "XMLHttpRequest",
            "Content-Type": "application/x-www-form-urlencoded",
        }

        res = self.session.http.post(
            self.token_url.format(deviceId=self.device_id, **config),
            data=pdata,
            headers=headers,
        )

        if res.status_code == 204:
            log.error(f"self._domain might be invalid - {self._domain}")
            return

        data = self.session.http.json(
            res,
            schema=validate.Schema({
                "token": str,
                validate.optional("name"): str,
                validate.optional("webPlayer"): {
                    validate.optional("id"): str,
                    validate.optional("name"): str,
                    validate.optional("type"): str,
                    validate.optional("allowedDomains"): [str],
                },
            }),
        )
        log.trace("%r", data)
        self.title = data.get("name")
        return data["token"]

    def _get_streams(self):
        if not self.matches["streann"]:
            self._domain = urlparse(self.url).netloc
            iframes = self.session.http.get(
                self.url,
                schema=validate.Schema(
                    validate.parse_html(),
                    validate.xml_findall(".//iframe[@src]"),
                    validate.filter(lambda elem: urlparse(elem.attrib.get("src")).netloc == "ott.streann.com"),
                ),
            )
            if not iframes:
                log.error("Could not find 'ott.streann.com' iframe")
                return
            self.url = iframes[0].attrib.get("src")

        if not self._domain and self.get_option("url"):
            self._domain = urlparse(self.get_option("url")).netloc

        if self._domain is None:
            log.error("Missing source URL, use --streann-url")
            return

        self.session.http.headers.update({"Referer": self.url})
        # Get the query string
        encrypted_data = urlparse(self.url).query
        data = base64.b64decode(encrypted_data)
        # and decrypt it
        passphrase = self.passphrase()
        if passphrase:
            log.debug("Found passphrase")
            params = decrypt_openssl(data, passphrase)
            config = parse_qsd(params.decode("utf8"))
            log.trace("config: %r", config)
            token = self.get_token(**config)
            if not token:
                return
            hls_url = self.stream_url.format(time=self.time, deviceId=self.device_id, token=token, **config)
            log.debug(f"URL={hls_url}")
            return HLSStream.parse_variant_playlist(self.session, hls_url, acceptable_status=(200, 403, 404, 500))


__plugin__ = Streann



"""
$description Global live broadcasting and live broadcast archiving social platform.
$url twitcasting.tv
$type live
$metadata id
"""

from __future__ import annotations

import hashlib
import re
import sys

from streamlink.buffers import RingBuffer
from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginargument, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.plugin.api.websocket import WebsocketClient
from streamlink.stream.hls import HLSStream, HLSStreamReader, HLSStreamWriter
from streamlink.stream.stream import Stream, StreamIO
from streamlink.utils.url import update_qsd


log = getLogger(__name__)


class TwitCastingHLSStreamWriter(HLSStreamWriter):
    _re_preroll_map_segment = re.compile(r"-\d+\.\w+$", re.IGNORECASE)

    def should_filter_segment(self, segment):
        return (
            "preroll-" in segment.uri
            or (segment.map and self._re_preroll_map_segment.search(segment.map.uri))
            or super().should_filter_segment(segment)
        )


class TwitCastingHLSStreamReader(HLSStreamReader):
    __writer__ = TwitCastingHLSStreamWriter


class TwitCastingHLSStream(HLSStream):
    __reader__ = TwitCastingHLSStreamReader


@pluginmatcher(
    re.compile(r"https?://twitcasting\.tv/(?P<channel>[^/]+)"),
)
@pluginargument(
    "password",
    sensitive=True,
    metavar="PASSWORD",
    help="Password for private Twitcasting streams.",
)
class TwitCasting(Plugin):
    _URL_API_STREAMSERVER = "https://twitcasting.tv/streamserver.php"

    # prefer websocket streams over HLS streams due to latency reasons
    _WEIGHTS = {
        "ws_main": sys.maxsize,
        "ws_mobilesource": sys.maxsize - 1,
        "ws_base": sys.maxsize - 2,
        "hls_high": sys.maxsize - 10,
        "hls_medium": sys.maxsize - 11,
        "hls_low": sys.maxsize - 12,
    }

    @classmethod
    def stream_weight(cls, stream: str) -> tuple[float, str]:
        return (cls._WEIGHTS[stream], "none") if stream in cls._WEIGHTS else super().stream_weight(stream)

    def _api_query_streamserver(self):
        return self.session.http.get(
            self._URL_API_STREAMSERVER,
            params={
                "target": self.match["channel"],
                "mode": "client",
                "player": "pc_web",
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    validate.optional("movie"): {
                        "id": int,
                        "live": bool,
                    },
                    validate.optional("llfmp4"): {
                        "streams": {
                            str: validate.url(),
                        },
                    },
                    validate.optional("tc-hls"): {
                        "streams": {
                            str: validate.url(),
                        },
                    },
                },
                validate.union_get("movie", "llfmp4", "tc-hls"),
            ),
        )

    def _get_streams_hls(self, streams, params=None):
        for name, url in streams.items():
            yield f"hls_{name}", TwitCastingHLSStream(self.session, url, params=params)

    def _get_streams_websocket(self, streams, params=None):
        for name, url in streams.items():
            yield f"ws_{name}", TwitCastingStream(self.session, url, params=params)

    def _get_streams(self):
        movie, websocket, hls = self._api_query_streamserver()
        if not movie or not movie.get("id") or not movie.get("live"):
            log.error(f"No live stream available for user {self.match['channel']}")
            return
        if not websocket and not hls:
            log.error("Unsupported stream type")
            return

        self.id = movie.get("id")

        params = {}
        if password := self.options.get("password"):
            params |= {"word": hashlib.md5(password.encode()).hexdigest()}

        if websocket:
            yield from self._get_streams_websocket(websocket["streams"], params)
        if hls:
            yield from self._get_streams_hls(hls["streams"], params)


class TwitCastingWsClient(WebsocketClient):
    def __init__(self, buffer: RingBuffer, *args, **kwargs):
        self.buffer = buffer
        super().__init__(*args, **kwargs)

    def on_close(self, *args, **kwargs):
        super().on_close(*args, **kwargs)
        self.buffer.close()

    def on_data(self, wsapp, data, data_type, cont):
        if data_type == self.OPCODE_TEXT:
            return

        try:
            self.buffer.write(data)
        except Exception as err:
            log.error(err)
            self.close()


class TwitCastingReader(StreamIO):
    def __init__(self, stream: TwitCastingStream, timeout=None):
        super().__init__()
        self.session = stream.session
        self.stream = stream
        self.timeout = timeout or self.session.options.get("stream-timeout")

        buffer_size = self.session.get_option("ringbuffer-size")
        self.buffer = RingBuffer(buffer_size)

        self.wsclient = TwitCastingWsClient(
            self.buffer,
            stream.session,
            stream.url,
            origin="https://twitcasting.tv/",
        )

    def open(self):
        self.wsclient.start()

    def close(self):
        self.wsclient.close()
        self.buffer.close()

    def read(self, size):
        return self.buffer.read(
            size,
            block=self.wsclient.is_alive(),
            timeout=self.timeout,
        )


class TwitCastingStream(Stream):
    __shortname__ = "websocket"

    def __init__(self, session, url, params):
        super().__init__(session)
        self.url = update_qsd(url, params or {})

    def to_url(self):
        return self.url

    def open(self):
        reader = TwitCastingReader(self)
        reader.open()
        return reader


__plugin__ = TwitCasting



from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, NamedTuple

from streamlink.logger import getLogger
from streamlink.stream.segmented.segment import Segment
from streamlink.utils.dataclass import FormattedDataclass
from streamlink.utils.l10n import Language


if TYPE_CHECKING:
    from datetime import datetime, timedelta


log = getLogger(".".join(__name__.split(".")[:-1]))


_MEDIA_LANGUAGE_CODES_RESERVED_LOCAL = re.compile(r"^q[a-t][a-z]$")
_MEDIA_LANGUAGE_CODES_PRIVATE_USE_SUBTAGS = re.compile(r"^[a-z]{2,3}-x-\S+$")


class Resolution(NamedTuple):
    width: int
    height: int


# EXTINF
class ExtInf(NamedTuple):
    duration: float  # version >= 3: float
    title: str | None


# EXT-X-BYTERANGE
class ByteRange(NamedTuple):  # version >= 4
    range: int
    offset: int | None


# EXT-X-DATERANGE
@dataclass(kw_only=True)
class DateRange(metaclass=FormattedDataclass):
    id: str | None
    classname: str | None
    start_date: datetime | None
    end_date: datetime | None
    duration: timedelta | None
    planned_duration: timedelta | None
    end_on_next: bool
    x: dict[str, str] = field(repr=False)


# EXT-X-KEY
@dataclass(kw_only=True)
class Key:
    method: str
    uri: str | None = field(repr=False)
    iv: bytes | None = field(repr=False)  # version >= 2
    key_format: str | None  # version >= 5
    key_format_versions: str | None  # version >= 5


# EXT-X-MAP
@dataclass(kw_only=True)
class Map:
    uri: str = field(repr=False)
    key: Key | None
    byterange: ByteRange | None


# EXT-X-MEDIA
@dataclass(kw_only=True)
class Media:
    uri: str | None
    type: str
    group_id: str
    language: str | None
    name: str
    default: bool
    autoselect: bool
    forced: bool
    characteristics: str | None

    parsed_language: Language | None = field(init=False, default=None, repr=False, hash=False, compare=False)

    def __post_init__(self):
        # parse the media playlist language, so we can compare it with the user's input
        self._parse_language()

    def _parse_language(self):
        if (
            self.language is None
            or _MEDIA_LANGUAGE_CODES_RESERVED_LOCAL.match(self.language)
            or _MEDIA_LANGUAGE_CODES_PRIVATE_USE_SUBTAGS.match(self.language)
        ):
            return

        try:
            self.parsed_language = Language.get(self.language)
        except LookupError:
            language = self.language
            name = self.name
            log.warning("Unrecognized language for media playlist: language=%r name=%r", language, name)


# EXT-X-START
class Start(NamedTuple):
    time_offset: float
    precise: bool


# EXT-X-STREAM-INF
@dataclass(kw_only=True)
class StreamInfo:
    bandwidth: int
    program_id: str | None  # version < 6
    codecs: list[str]
    resolution: Resolution | None
    framerate: float | None
    audio: str | None
    video: str | None
    subtitles: str | None


# EXT-X-I-FRAME-STREAM-INF
@dataclass(kw_only=True)
class IFrameStreamInfo:
    bandwidth: int
    program_id: str | None
    codecs: list[str]
    resolution: Resolution | None
    video: str | None


@dataclass(kw_only=True)
class HLSPlaylist:
    uri: str
    stream_info: StreamInfo | IFrameStreamInfo
    media: list[Media]
    is_iframe: bool


@dataclass(kw_only=True)
class HLSSegment(Segment):
    title: str | None
    key: Key | None
    byterange: ByteRange | None
    date: datetime | None
    map: Map | None
